import pandas as pd
import numpy as np
from pathlib import Path
from taxontabletools2.utilities import filter_taxontable
import plotly.graph_objects as go
import plotly.express as px

def read_props_bar(path_to_outdirs, taxon_table_xlsx, taxon_table_df, samples, metadata_df, selected_metadata,
                   traits_df, selected_traits, user_settings, tool_settings):

    # Make copies of input dataframes to prevent altering the originals
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    # Extract relevant settings from the tool settings
    selected_metadata = tool_settings['selected_metadata']
    taxonomic_level = tool_settings['taxonomic_level']

    # Get unique taxa at the specified taxonomic level and exclude empty values
    all_taxa = taxon_table_df[taxonomic_level].dropna().unique()

    # Initialize result dictionary
    res = {}

    if selected_metadata == 'All samples':
        # Case where all samples are selected (simpler computation)
        total_reads_per_sample = taxon_table_df[samples].sum(axis=0)

        for taxon in all_taxa:
            # Filter taxon-specific data only once per taxon
            sub_df = taxon_table_df[taxon_table_df[taxonomic_level] == taxon]

            # Sum reads per taxon for each sample and calculate the proportion
            taxon_reads_per_sample = sub_df[samples].sum(axis=0)
            read_props = (taxon_reads_per_sample / total_reads_per_sample * 100).fillna(0).values
            res[taxon] = read_props
        res_df = pd.DataFrame(res, index=samples).transpose()

    else:
        # Case where specific categories from metadata are selected
        categories = metadata_df[selected_metadata].dropna().unique()

        # Filter metadata to keep relevant columns only
        metadata_df = metadata_df[['Sample', selected_metadata]]

        for taxon in all_taxa:
            sub_df = taxon_table_df[taxon_table_df[taxonomic_level] == taxon]

            reads_list = []
            for category in categories:
                # Get the samples belonging to the current category
                category_samples = metadata_df.loc[metadata_df[selected_metadata] == category, 'Sample'].tolist()

                # Calculate total and taxon-specific reads for the category
                total_reads = taxon_table_df[category_samples].sum().sum()
                taxon_reads = sub_df[category_samples].sum().sum()

                # Compute proportion
                read_props = (taxon_reads / total_reads * 100) if total_reads != 0 else 0
                reads_list.append(read_props)

            res[taxon] = reads_list
        res_df = pd.DataFrame(res, index=categories).transpose()

    # Use color map based on user settings
    colorscale_name = user_settings.get('colorsequence', 'Plotly')
    colors = getattr(px.colors.qualitative, colorscale_name, px.colors.qualitative.Plotly)

    # Initialize and populate the figure
    fig = go.Figure()

    for i, taxon in enumerate(res_df.index):
        fig.add_trace(go.Bar(
            x=res_df.columns,
            y=res_df.loc[taxon],
            name=taxon,
            hoverinfo='y+name',
            marker_color=colors[i % len(colors)]
        ))

    # Update layout
    fig.update_xaxes(dtick='linear')
    fig.update_yaxes(range=(0, 100))
    fig.update_layout(barmode='stack',
                      title=f'Proportion of Reads: {taxonomic_level}, {selected_metadata}',
                      template=user_settings['template'],
                      font_size=user_settings['font_size'],
                      yaxis_title='Proportion of Reads (%)',
                      yaxis_title_font=dict(size=user_settings['font_size']),
                      yaxis_tickfont=dict(size=user_settings['font_size']),
                      xaxis_title_font=dict(size=user_settings['font_size']),
                      xaxis_tickfont=dict(size=user_settings['font_size']))

    return fig


def read_props_heatmap(path_to_outdirs, taxon_table_xlsx, taxon_table_df, samples, metadata_df, selected_metadata,
                       traits_df, selected_traits, user_settings, tool_settings):

    # Make copies of input dataframes to prevent altering the originals
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    # Extract relevant settings from the tool settings
    selected_metadata = tool_settings['selected_metadata']
    taxonomic_level = tool_settings['taxonomic_level']

    # Get unique taxa at the specified taxonomic level and exclude empty values
    all_taxa = taxon_table_df[taxonomic_level].dropna().unique()

    # Initialize result dictionary
    res = {}

    if selected_metadata == 'All samples':
        # Case where all samples are selected (simpler computation)
        total_reads_per_sample = taxon_table_df[samples].sum(axis=0)

        for taxon in all_taxa:
            # Filter taxon-specific data only once per taxon
            sub_df = taxon_table_df[taxon_table_df[taxonomic_level] == taxon]

            # Sum reads per taxon for each sample and calculate the proportion
            taxon_reads_per_sample = sub_df[samples].sum(axis=0)
            read_props = (taxon_reads_per_sample / total_reads_per_sample * 100).fillna(0).values
            res[taxon] = read_props
        res_df = pd.DataFrame(res, index=samples).transpose()

    else:
        # Case where specific categories from metadata are selected
        categories = [i for i in metadata_df[selected_metadata].dropna().unique() if i != '']

        # Filter metadata to keep relevant columns only
        metadata_df = metadata_df[['Sample', selected_metadata]]

        for taxon in all_taxa:
            sub_df = taxon_table_df[taxon_table_df[taxonomic_level] == taxon]

            reads_list = []
            for category in categories:
                # Get the samples belonging to the current category
                category_samples = metadata_df.loc[metadata_df[selected_metadata] == category, 'Sample'].tolist()

                # Calculate total and taxon-specific reads for the category
                total_reads = taxon_table_df[category_samples].sum().sum()
                taxon_reads = sub_df[category_samples].sum().sum()

                # Compute proportion
                read_props = (taxon_reads / total_reads * 100) if total_reads != 0 else 0
                reads_list.append(read_props)

            res[taxon] = reads_list
        res_df = pd.DataFrame(res, index=categories).transpose()

    # Use color map based on user settings
    colorscale_name = user_settings.get('colorscale', 'Viridis')

    # Initialize heatmap figure
    fig = go.Figure(data=go.Heatmap(
        z=res_df.values.tolist(),
        x=res_df.columns.tolist(),
        y=res_df.index.tolist(),
        hoverongaps=False,
        colorscale=colorscale_name))

    # Update layout
    fig.update_xaxes(dtick='linear')
    fig.update_yaxes(dtick='linear')
    fig.update_layout(
        title=f'Proportion of Reads: {taxonomic_level}, {selected_metadata}',
        template=user_settings['template'],
        font_size=user_settings['font_size'],
        yaxis_title='Proportion of Reads (%)',
        yaxis_title_font=dict(size=user_settings['font_size']),
        yaxis_tickfont=dict(size=user_settings['font_size']),
        xaxis_title_font=dict(size=user_settings['font_size']),
        xaxis_tickfont=dict(size=user_settings['font_size']))

    return fig
















import pandas as pd
import numpy as np
from pathlib import Path
from taxontabletools2.utilities import filter_taxontable
import plotly.graph_objects as go
import plotly.express as px
from scipy.spatial.distance import pdist, squareform
from scipy.spatial import ConvexHull
import streamlit as st
import plotly.colors

def pcoa_analysis(path_to_outdirs, taxon_table_xlsx, taxon_table_df, samples, metadata_df, selected_metadata,
                    traits_df, selected_traits, user_settings, tool_settings):

    # Make copies of input dataframes to prevent altering the originals
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    # Use order of samples as sorted in the metadata_df
    samples = metadata_df['Sample'].values.tolist()

    # Extract relevant settings from the tool settings
    selected_metadata = tool_settings['selected_metadata']
    taxonomic_level = tool_settings['taxonomic_level']
    metric = tool_settings['metric']

    if selected_metadata != 'All samples':

        ## Select only the relevant samples
        sub_df = metadata_df[['Sample', selected_metadata]]
        sub_df = sub_df.loc[sub_df[selected_metadata] != '']
        samples = sub_df['Sample'].values.tolist()
        sample_categories = sub_df[selected_metadata].values.tolist()

        # Create colors for categories
        colorscale_name = user_settings.get('colorsequence', 'Plotly')
        colors = getattr(px.colors.qualitative, colorscale_name, px.colors.qualitative.Plotly)*200
        categories = sub_df[selected_metadata].drop_duplicates().values.tolist()
        color_dict = {i:j for i,j in zip(categories, colors)}
        sample_colors = [color_dict[metadata] for sample, metadata in sub_df.values.tolist()]

    else:
        # Create a color for each sample (use continous colors here)
        sample_categories = samples
        num_colors = len(sample_categories)
        sample_colors = plotly.colors.sample_colorscale(user_settings['colorscale'], [i / (num_colors - 1) for i in range(num_colors)])

    # Get unique taxa at the specified taxonomic level and exclude empty values
    all_taxa = [i for i in taxon_table_df[taxonomic_level].dropna().unique() if i != '']

    # Initialize result dictionary
    res = {}

    # Case where all samples are used
    for sample in samples:
        # Sum the reads for each taxon in the current sample and filter for non-zero values
        sample_taxa = taxon_table_df[[taxonomic_level, sample]].groupby(taxonomic_level).sum()
        sample_taxa = sample_taxa.loc[sample_taxa[sample] > 0, sample]

        # Store binary presence/absence data if metric is 'Jaccard', otherwise store actual values
        if metric == 'Jaccard':
            res[sample] = [1 if taxon in sample_taxa.index else 0 for taxon in all_taxa]
        else:
            res[sample] = [sample_taxa[taxon] if taxon in sample_taxa.index else 0 for taxon in all_taxa]

    # Convert result dict into DataFrame (rows: taxa, columns: samples)
    presence_absence_df = pd.DataFrame(res, index=all_taxa).transpose()

    # Calculate distance between samples
    if metric == 'Jaccard':
        distances = pdist(presence_absence_df, metric=metric)
    else:
        distances = pdist(presence_absence_df, metric=metric)

    # Convert to square form distance matrix
    distance_matrix = squareform(distances)

    # Convert to a DataFrame for easier manipulation
    distance_matrix_df = pd.DataFrame(distance_matrix, index=samples, columns=samples)

    # Try to import skbio (which is difficult to install)
    try:
        from skbio.stats.ordination import pcoa
        from skbio import DistanceMatrix
        from skbio.stats.distance import anosim
        skbio = True
    except ImportError:
        skbio = False

    # Only continue if skbio is installed
    if skbio == True:

        # Calculate pcoa
        pcoa_res = pcoa(distance_matrix_df)

        # Collect expained variance
        pcoa_explained_variance_df = pd.DataFrame(pcoa_res.proportion_explained, columns=['explained_variance'])*100
        pcoa_explained_variance_df = pcoa_explained_variance_df[pcoa_explained_variance_df['explained_variance'] > 1]

        # Collect values and already filter for PC axes with >1% explained variance
        pcoa_df = pd.DataFrame(pcoa_res.samples)
        pcoa_df.index=samples
        pcoa_axes = list(pcoa_explained_variance_df.index)
        pcoa_df = pcoa_df[pcoa_axes]
        pcoa_df['Color'] = sample_colors
        pcoa_df['Metadata'] = sample_categories

        # Create a dict for display and to draw from later
        axes_dict = pcoa_explained_variance_df.to_dict()['explained_variance']

        # Calculate anosim
        # Calculate ANOSIM if specific metadata is selected
        anosim_result = None
        if selected_metadata != 'All samples':
            # Create DistanceMatrix object for ANOSIM
            dm = DistanceMatrix(distance_matrix)

            # Perform ANOSIM using the selected metadata as grouping variable
            result = anosim(dm, sample_categories, permutations=999)
            anosim_result = result  # Store the ANOSIM result

        return pcoa_df, axes_dict, anosim_result, selected_metadata

    else:
        st.error('Please install skbio to calculate PCoAs!')

def draw_outlines(fig, x_values, y_values, metadata, color):
    ## collect samples that form the outline
    x_plane, y_plane = [], []
    hull = ConvexHull(np.column_stack((x_values, y_values)))
    for i in hull.vertices:
        x_plane.append(x_values[i])
        y_plane.append(y_values[i])
    x_plane.append(x_values[hull.vertices[0]])
    y_plane.append(y_values[hull.vertices[0]])

    ## draw the outline
    fig.add_trace(go.Scatter(x=x_plane, y=y_plane, mode='lines', name=metadata, marker_color=color, fill='toself'))

def display_pcoa(pcoa_df, axes_dict, selected_axes, user_settings, tool_settings):

        # Extract axes to display
        display_axes = [i.split(' ')[0] for i in selected_axes if i != 'Not selected']
        explained_variances = {i:axes_dict[i] for i in display_axes}
        pcoa_axes = list(explained_variances.keys())
        pcoa_df_selected_axes = pcoa_df[pcoa_axes]

        ## Display anosim
        anosim_result = tool_settings['anosim_result']
        if anosim_result is not None:
            r_value = anosim_result['test statistic']
            p_value = anosim_result['p-value']
            title = f'{tool_settings["pcoa_metadata"]}; {tool_settings["taxonomic_level"]}; ANOSIM R={r_value}, p={p_value}'
        else:
            title = f'{tool_settings["pcoa_metadata"]}; {tool_settings["taxonomic_level"]}; ANOSIM not calculated'

        if len(explained_variances) == 2:
            pcoa_2D_df = pcoa_df_selected_axes.copy()
            pcoa_2D_df['Color'] = pcoa_df['Color'].values.tolist()
            pcoa_2D_df['Metadata'] = pcoa_df['Metadata'].values.tolist()

            fig = go.Figure()

            for metadata in set(pcoa_df['Metadata'].values.tolist()):
                sub_df = pcoa_2D_df.loc[pcoa_2D_df['Metadata'] == metadata]

                x_values = sub_df[pcoa_axes[0]].values.tolist()
                y_values = sub_df[pcoa_axes[1]].values.tolist()
                text_values = list(sub_df.index)
                colors = sub_df['Color'].values.tolist()

                fig.add_trace(go.Scatter(x=x_values, y=y_values, text=text_values, marker_color=colors, marker=dict(size=user_settings['scatter_size']),  name=metadata, mode='markers'))

            if tool_settings['pcoa_metadata'] != 'All samples' and tool_settings['draw_outlines'] == True:
                for category in set(pcoa_df['Metadata'].values.tolist()):
                    sub_df = pcoa_df.loc[pcoa_df['Metadata'] == category]
                    x_values = sub_df[pcoa_axes[0]].values.tolist()
                    y_values = sub_df[pcoa_axes[1]].values.tolist()
                    color = sub_df['Color'].drop_duplicates().values.tolist()[0]
                    draw_outlines(fig, x_values, y_values, category, color)

            # Update layout
            fig.update_xaxes(title=f'{selected_axes[0]}')
            fig.update_yaxes(title=f'{selected_axes[1]}')
            fig.update_layout(title=title,
                              template=user_settings['template'],
                              font_size=user_settings['font_size'],
                              width=user_settings['plot_width'],
                              height=user_settings['plot_height'],
                              showlegend=user_settings['show_legend'],
                              yaxis_title_font=dict(size=user_settings['font_size']),
                              yaxis_tickfont=dict(size=user_settings['font_size']),
                              xaxis_title_font=dict(size=user_settings['font_size']),
                              xaxis_tickfont=dict(size=user_settings['font_size']))

            return fig

        elif len(explained_variances) == 3:

            pcoa_3D_df = pcoa_df_selected_axes.copy()
            pcoa_3D_df['Color'] = pcoa_df['Color'].values.tolist()
            pcoa_3D_df['Metadata'] = pcoa_df['Metadata'].values.tolist()

            fig = go.Figure()

            for metadata in set(pcoa_df['Metadata'].values.tolist()):
                sub_df = pcoa_3D_df.loc[pcoa_3D_df['Metadata'] == metadata]

                x_values = sub_df[pcoa_axes[0]].values.tolist()
                y_values = sub_df[pcoa_axes[1]].values.tolist()
                z_values = sub_df[pcoa_axes[2]].values.tolist()
                text_values = list(sub_df.index)
                colors = sub_df['Color'].values.tolist()

                fig.add_trace(go.Scatter3d(x=x_values, y=y_values, z=z_values, text=text_values, marker_color=colors, marker=dict(size=user_settings['scatter_size']),  name=metadata, mode='markers'))

            # Update layout with custom axis titles
            fig.update_layout(
                scene=dict(
                    xaxis_title=f'{selected_axes[0]}',  # Replace with your X axis title
                    yaxis_title=f'{selected_axes[1]}',  # Replace with your Y axis title
                    zaxis_title=f'{selected_axes[2]}'  # Replace with your Z axis title
                )
            )

            fig.update_layout(title=title,
                              template=user_settings['template'],
                              font_size=user_settings['font_size'],
                              width=user_settings['plot_width'],
                              height=user_settings['plot_height'],
                              showlegend=user_settings['show_legend'],
                              yaxis_title_font=dict(size=user_settings['font_size']),
                              yaxis_tickfont=dict(size=user_settings['font_size']),
                              xaxis_title_font=dict(size=user_settings['font_size']),
                              xaxis_tickfont=dict(size=user_settings['font_size']))

            return fig









import streamlit as st
import pandas as pd
import numpy as np
import subprocess
import glob, sys, time, statistics, os.path, random
from pathlib import Path
import psutil
import plotly.graph_objects as go
from taxontabletools2.start import TTT_variables
from taxontabletools2.utilities import load_df
from taxontabletools2.utilities import collect_replicates
from taxontabletools2.utilities import collect_sample_stats
from taxontabletools2.basic_stats import basic_stats_reads
from taxontabletools2.basic_stats import basic_stats_OTUs
from taxontabletools2.basic_stats import taxonomic_resolution
from taxontabletools2.basic_stats import taxonomic_richness
from taxontabletools2.utilities import collect_traits
from taxontabletools2.utilities import strip_traits
from taxontabletools2.taxontable_conversion import taxon_table_converter_ttt
from taxontabletools2.taxontable_conversion import taxon_table_converter_qiime2
import streamlit as st
from taxontabletools2.start import start

st.session_state['config'] = {
    'toImageButtonOptions': {'format': 'svg', 'filename': 'custom_TTT_image'},
    'modeBarButtonsToRemove': ['zoom2d', 'pan2d', 'lasso2d', 'zoomIn2d', 'zoomOut2d', 'autoScale2d', 'hoverClosestCartesian', 'hoverCompareCartesian'],
    'modeBarButtonsToAdd': ['drawline', 'drawopenpath', 'drawclosedpath', 'drawcircle', 'drawrect', 'eraseshape']
}

st.set_page_config(layout="wide")

# Call the sidebar function
settings = start()
users_settings = settings['new_user_preferences_dict']

# Ensure the session state variables are retained across subpages
if 'TaXon_table_df' not in st.session_state:
    error1, error2, error3 = st.columns(3)
    with error2:
        st.write('##### Please load a TaXon table to continue!')

elif 'TaXon_table_xlsx' not in st.session_state:
    error1, error2, error3 = st.columns(3)
    with error2:
        st.write('##### Please load a TaXon table to continue!')

elif st.session_state['TaXon_table_df'].empty == True:
    error1, error2, error3 = st.columns(3)
    with error2:
        st.write('##### Please check your TaXon table. It seems to be empty!')

elif st.session_state['TaXon_table_xlsx'].is_file() == False:
    error1, error2, error3 = st.columns(3)
    with error2:
        st.write('##### The TaXon table does not exist in the selected project folder!')

else:
    # Display the taxon table if loaded
    st.write("""# TaxonTableTools 2.0""")

    if not st.session_state['TaXon_table_df'].empty:
        if st.session_state['TaXon_table_xlsx'].is_file() != False:
            st.write(""" ## Basic statistics """)

            ## collect stats
            st.session_state['stats_df'] = collect_sample_stats(st.session_state['TaXon_table_df'],
                                                                st.session_state['TaXon_table_xlsx'],
                                                                st.session_state['traits_df'],
                                                                st.session_state['samples'],
                                                                st.session_state['metadata_df'])

            ## PLOTS
            a1, a2 = st.columns(2)
            with a1:
                ## calculate read numbers
                fig_a1, min_reads, max_reads, avg_reads, stdev_reads = basic_stats_reads(users_settings, st.session_state['TaXon_table_df'])
                ## plot reads
                st.plotly_chart(fig_a1, use_container_width=True, config=st.session_state['config'])
                st.write('The samples contain between **{}** and **{}** reads with an average of **{} (+-{})**.'.format(
                    min_reads, max_reads, avg_reads, stdev_reads))

            with a2:
                ## calculate number of OTUs
                fig_a2, min_OTUs, max_OTUs, avg_OTUs = basic_stats_OTUs(users_settings, st.session_state['TaXon_table_df'])
                ## plot number of OTUs
                st.plotly_chart(fig_a2, use_container_width=True, config=st.session_state['config'])
                st.write('The samples contain between **{}** and **{}** {} with an average of **{}**.'.format(min_OTUs,
                                                                                                              max_OTUs,
                                                                                                              users_settings['clustering_unit'],
                                                                                                              avg_OTUs))

            a3, a4 = st.columns(2)
            with a3:
                ## calculate read numbers
                fig_a3 = taxonomic_richness(users_settings, st.session_state['TaXon_table_df'])
                ## plot reads
                st.plotly_chart(fig_a3, use_container_width=True, config=st.session_state['config'])
                st.write('Taxonomic richness')

            with a4:
                ## calculate number of OTUs
                fig_a4 = taxonomic_resolution(users_settings, st.session_state['TaXon_table_df'])
                ## plot number of OTUs
                st.plotly_chart(fig_a4, use_container_width=True, config=st.session_state['config'])
                st.write('Taxonomic resolution')

            ## Per sample information table and Metadata table
            a5, a6 = st.columns(2)
            with a5:
                st.write(""" ## Per sample information """)
                st.table(st.session_state['stats_df'])
            with a6:
                st.write(""" ## Available metadata """)
                st.table(st.session_state['metadata_df'])

        else:
            error1, error2, error3 = st.columns(3)
            with error2:
                st.write(
                    '##### Could not find your TaXon table in the selected project. Please make sure to select the according project!')

    else:
        st.error('TaXon table is empty. Please load a table to continue.')







import streamlit as st
import pandas as pd
import os
import sys
import subprocess
import webbrowser
import glob
from pathlib import Path
from taxontabletools2.utilities import load_df, collect_traits, strip_traits
import plotly.express as px

def TTT_variables():

    directories_to_create = ["Venn_diagrams","TaXon_tables", "TaXon_tables_per_sample", "Taxon_lists", "Rarefaction_curves",
    "Taxonomic_richness_plots", "CCA_plots", "Taxonomic_resolution_plots", "Meta_data_table", "Site_occupancy_plots", "Read_proportions_plots", "Basic_stats",
    "Krona_charts", "Perlodes", "Alpha_diversity", "Beta_diversity", "PCoA_plots", "Replicate_analysis", "GBIF", "Occurrence_analysis", "Per_taxon_statistics",
    "ParCat_plots", "NMDS_plots", "Table_comparison"]

    available_templates_list = ['seaborn', 'ggplot2', 'simple_white', 'plotly', 'plotly_dark', 'presentation', 'plotly_white']

    available_clustering_units = ['OTUs', 'zOTUs', 'ESVs', 'ASVs']

    plotly_colors = ["aliceblue", "antiquewhite", "aqua", "aquamarine", "azure",
    "beige", "bisque", "black", "blanchedalmond", "blue",
    "blueviolet", "brown", "burlywood", "cadetblue",
    "chartreuse", "chocolate", "coral", "cornflowerblue",
    "cornsilk", "crimson", "cyan", "darkblue", "darkcyan",
    "darkgoldenrod", "darkgray", "darkgrey", "darkgreen",
    "darkkhaki", "darkmagenta", "darkolivegreen", "darkorange",
    "darkorchid", "darkred", "darksalmon", "darkseagreen",
    "darkslateblue", "darkslategray", "darkslategrey",
    "darkturquoise", "darkviolet", "deeppink", "deepskyblue",
    "dimgray", "dimgrey", "dodgerblue", "firebrick",
    "floralwhite", "forestgreen", "fuchsia", "gainsboro",
    "ghostwhite", "gold", "goldenrod", "gray", "grey", "green",
    "greenyellow", "honeydew", "hotpink", "indianred", "indigo",
    "ivory", "khaki", "lavender", "lavenderblush", "lawngreen",
    "lemonchiffon", "lightblue", "lightcoral", "lightcyan",
    "lightgoldenrodyellow", "lightgray", "lightgrey",
    "lightgreen", "lightpink", "lightsalmon", "lightseagreen",
    "lightskyblue", "lightslategray", "lightslategrey",
    "lightsteelblue", "lightyellow", "lime", "limegreen",
    "linen", "magenta", "maroon", "mediumaquamarine",
    "mediumblue", "mediumorchid", "mediumpurple",
    "mediumseagreen", "mediumslateblue", "mediumspringgreen",
    "mediumturquoise", "mediumvioletred", "midnightblue",
    "mintcream", "mistyrose", "moccasin", "navajowhite", "navy",
    "oldlace", "olive", "olivedrab", "orange", "orangered",
    "orchid", "palegoldenrod", "palegreen", "paleturquoise",
    "palevioletred", "papayawhip", "peachpuff", "peru", "pink",
    "plum", "powderblue", "purple", "red", "rosybrown",
    "royalblue", "saddlebrown", "salmon", "sandybrown",
    "seagreen", "seashell", "sienna", "silver", "skyblue",
    "slateblue", "slategray", "slategrey", "snow", "springgreen",
    "steelblue", "tan", "teal", "thistle", "tomato", "turquoise",
    "violet", "wheat", "white", "whitesmoke", "yellow",
    "yellowgreen"]

    available_colorsequences = {"Plotly":px.colors.qualitative.Plotly, "G10":px.colors.qualitative.G10,
    "T10":px.colors.qualitative.T10, "Alphabet":px.colors.qualitative.Alphabet, "Dark24":px.colors.qualitative.Dark24, "Dark24_r":px.colors.qualitative.Dark24_r,
    "Light24":px.colors.qualitative.Light24, "Set1":px.colors.qualitative.Set1, "Pastel1":px.colors.qualitative.Pastel,
    "Dark2":px.colors.qualitative.Dark2, "Set2":px.colors.qualitative.Set2, "Pastel2":px.colors.qualitative.Pastel2,
    "Set3":px.colors.qualitative.Set3, "Antique":px.colors.qualitative.Antique,"Bold":px.colors.qualitative.Bold,
    "Pastel":px.colors.qualitative.Pastel, "Prism":px.colors.qualitative.Prism, "Safe":px.colors.qualitative.Safe,
    "Vivid":px.colors.qualitative.Vivid}

    available_colorscales = px.colors.named_colorscales()

    available_taxonomic_levels_list= ['Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species', 'ID']
    st.session_state['available_taxonomic_levels_list'] = available_taxonomic_levels_list

    return [directories_to_create, available_templates_list, available_clustering_units, plotly_colors, available_colorsequences, available_colorscales, available_taxonomic_levels_list]

def start():
    # Collect path to default files
    path_to_ttt = Path(__file__).resolve().parent
    user_preferences_xlsx = path_to_ttt.joinpath('user_preferences.xlsx')
    user_preferences_df = pd.read_excel(user_preferences_xlsx)
    path_to_projects = Path(user_preferences_df.loc[user_preferences_df['Variable'] == 'path_to_outdirs', 'Value'].values[0])
    st.session_state['perlodes_TTT_conversion_xlsx'] = path_to_ttt.joinpath('WFD_conversion/perlodes_TTT_conversion.xlsx')

    projects_dict = {}
    for project in sorted(glob.glob(str(path_to_projects.joinpath('*')))):
        projects_dict[Path(project).stem] = Path(project)

    # If this is the first load, initialize the session state values.
    if 'TaXon_table_df' not in st.session_state:
        st.session_state['TaXon_table_df'] = pd.DataFrame()

    if 'project_name' not in st.session_state:
        st.session_state['project_name'] = list(projects_dict.keys())[0]

    directories_to_create, available_templates_list, available_clustering_units, plotly_colors, available_colorsequences, available_colorscales, available_taxonomic_levels_list = TTT_variables()

    user_preferences_dict = {key: value for (key, value) in user_preferences_df.values.tolist()}
    default_template = user_preferences_dict['template']
    default_height = user_preferences_dict['plot_height']
    default_width = user_preferences_dict['plot_width']
    default_show_legend = user_preferences_dict['show_legend']
    default_color1 = user_preferences_dict['color_1']
    default_color2 = user_preferences_dict['color_2']
    default_colorsequence = user_preferences_dict['colorsequence']
    default_colorscale = user_preferences_dict['colorscale']
    default_font_size = int(user_preferences_dict['font_size'])
    default_clustering_unit = user_preferences_dict['clustering_unit']

    # Create empty df to store intermediate TaXon tables
    cache_df = pd.DataFrame()
    TaXon_table_df = pd.DataFrame()

    ### Sidebar

    st.sidebar.write(""" # Project """)
    st.sidebar.write(""" ### TaXon Table """)

    ## CALL PROJECTS
    available_projects_list = list(projects_dict.keys())
    if 'project_name' not in st.session_state:
        st.session_state['project_name'] = st.sidebar.selectbox('Select a project:', available_projects_list)
        st.session_state['path_to_outdirs'] = projects_dict[st.session_state['project_name']]
    else:
        value = st.session_state['project_name']
        st.session_state['project_name'] = st.sidebar.selectbox('Select a project:', available_projects_list, index=available_projects_list.index(value))
        st.session_state['path_to_outdirs'] = projects_dict[st.session_state['project_name']]

    ## CALL PROJECTS
    if st.sidebar.button('Reveal project in finder'):
        if sys.platform == 'win32':
            subprocess.Popen(['start', st.session_state['path_to_outdirs']], shell=True)
        elif sys.platform == 'darwin':
            subprocess.Popen(['open', st.session_state['path_to_outdirs']])

    if st.sidebar.button('Open table'):
        pass

    ## load TaXon table
    TaXon_table_upload = st.sidebar.file_uploader('Load TaXon table')

    if 'TaXon_table_df' not in st.session_state:
        st.session_state['TaXon_table_df'] = pd.DataFrame()

    if st.sidebar.button('Load table'):
        if TaXon_table_upload is not None:
            # Reset the session state before loading new data
            st.session_state['TaXon_table_df'] = pd.DataFrame()  # Clear the old table

            # Load the new table
            TaXon_table_xlsx = Path(
                '{}/TaXon_tables/{}'.format(st.session_state['path_to_outdirs'], TaXon_table_upload.name))
            st.session_state['TaXon_table_xlsx'] = TaXon_table_xlsx

            # Verify that the file exists before attempting to load it
            if not st.session_state['TaXon_table_xlsx'].is_file():
                error1, error2, error3 = st.columns(3)
                with error2:
                    st.write('##### The TaXon table does not exist in the selected project folder!')
            else:
                # Load the TaXon table
                TaXon_table_df = load_df(st.session_state['TaXon_table_xlsx'])

                # Update session state with the new table
                st.session_state['TaXon_table_df'] = TaXon_table_df

                # Process the traits and other data
                traits_df = collect_traits(st.session_state['TaXon_table_df'])
                st.session_state['traits_df'] = traits_df
                TaXon_table_df = strip_traits(st.session_state['TaXon_table_df'])
                st.session_state['TaXon_table_df'] = TaXon_table_df
                samples = list(st.session_state['TaXon_table_df'].columns[9:])
                st.session_state['samples'] = samples

                # Load the metadata if it exists
                try:
                    metadata_df = pd.read_excel(st.session_state['TaXon_table_xlsx'], sheet_name='Metadata Table').fillna('')
                    st.session_state['metadata_df'] = metadata_df
                    if sorted(st.session_state['metadata_df']['Sample'].values.tolist()) != sorted(st.session_state['samples']):
                        st.session_state['metadata_df'] = pd.DataFrame(columns=['Sample', 'Metadata'])
                except ValueError:
                    dummy_metadata = [[j,i] for i,j in enumerate(samples)]
                    st.session_state['metadata_df'] = pd.DataFrame(dummy_metadata, columns=['Sample', 'Placeholder'])

                if sorted(samples) != sorted(st.session_state['metadata_df']['Sample'].values.tolist()):
                    dummy_metadata = [[j,i] for i,j in enumerate(samples)]
                    st.session_state['metadata_df'] = pd.DataFrame(dummy_metadata, columns=['Sample', 'Placeholder'])
                    st.error('Warning: The samples in the metadata sheet do not match with the samples in the taxontable sheet!')


        else:
            st.warning('Please provide a Taxon Table file first!')

    if 'TaXon_table_xlsx' in st.session_state:
        st.sidebar.write(Path(st.session_state['TaXon_table_xlsx']).name)

    ####################################################################################################################

    st.sidebar.write(""" # Settings """)
    st.sidebar.write(""" ### Presentation """)

    # PLOT HEIGHT
    if 'plot_height' not in st.session_state:
        st.session_state['plot_height'] = st.sidebar.number_input('Height', 400, 4000, default_height)
    else:
        value = st.session_state['plot_height']
        st.session_state['plot_height'] = st.sidebar.number_input('Height', 400, 4000, value)

    # PLOT WIDTH
    if 'plot_width' not in st.session_state:
        st.session_state['plot_width'] = st.sidebar.number_input('Width', 400, 4000, default_width)
    else:
        value = st.session_state['plot_width']
        st.session_state['plot_width'] = st.sidebar.number_input('Width', 400, 4000, value)

    # SHOW LEGEND
    if 'show_legend' not in st.session_state:
        st.session_state['show_legend'] = st.sidebar.selectbox('Show legend', [True, False], index=[True, False].index(default_show_legend))
    else:
        value = st.session_state['show_legend']
        st.session_state['show_legend'] = st.sidebar.selectbox('Show legend', [True, False], index=[True, False].index(value))

    # TEMPLATE
    if 'template' not in st.session_state:
        st.session_state['template'] = st.sidebar.selectbox('Layout', available_templates_list, index=available_templates_list.index(default_template))
    else:
        value = st.session_state['template']
        st.session_state['template'] = st.sidebar.selectbox('Layout', available_templates_list,
                                                            index=available_templates_list.index(value))

    # FONT_SIZE
    if 'font_size' not in st.session_state:
        st.session_state['font_size'] = st.sidebar.slider('Font size', 6, 30, default_font_size)
    else:
        value = st.session_state['font_size']
        st.session_state['font_size'] = st.sidebar.slider('Font size', 6, 30, value)

    # CLUSTERING UNIT
    if 'clustering_unit' not in st.session_state:
        st.session_state['clustering_unit'] = st.sidebar.selectbox('Clustering Unit', available_clustering_units,
                                                                   index=available_clustering_units.index(
                                                                       default_clustering_unit))
    else:
        value = st.session_state['clustering_unit']
        st.session_state['clustering_unit'] = st.sidebar.selectbox('Clustering Unit', available_clustering_units,
                                                                   index=available_clustering_units.index(value))

    # SCATTER SIZE
    if 'scatter_size' not in st.session_state:
        st.session_state['scatter_size'] = st.sidebar.number_input('Scatter size', 0, 40, 15)
    else:
        value = st.session_state['scatter_size']
        st.session_state['scatter_size'] = st.sidebar.number_input('Scatter size', 0, 40, value)

    st.sidebar.write(""" ### Colors """)

    # COLOR 1
    if 'color_1' not in st.session_state:
        st.session_state['color_1'] = st.sidebar.selectbox('Color 1', plotly_colors,
                                                           index=plotly_colors.index(default_color1))
    else:
        value = st.session_state['color_1']
        st.session_state['color_1'] = st.sidebar.selectbox('Color 1', plotly_colors, index=plotly_colors.index(value))

    # COLOR 2
    if 'color_2' not in st.session_state:
        st.session_state['color_2'] = st.sidebar.selectbox('Color 2', plotly_colors,
                                                           index=plotly_colors.index(default_color2))
    else:
        value = st.session_state['color_2']
        st.session_state['color_2'] = st.sidebar.selectbox('Color 2', plotly_colors, index=plotly_colors.index(value))

    # COLORSEQUENCE
    if 'colorsequence' not in st.session_state:
        st.session_state['colorsequence'] = st.sidebar.selectbox('Color sequence',
                                                                 list(available_colorsequences.keys()),
                                                                 index=list(available_colorsequences.keys()).index(
                                                                     default_colorsequence))
    else:
        value = st.session_state['colorsequence']
        st.session_state['colorsequence'] = st.sidebar.selectbox('Color sequence',
                                                                 list(available_colorsequences.keys()),
                                                                 index=list(available_colorsequences.keys()).index(
                                                                     value))

    # COLORSCALE
    if 'colorscale' not in st.session_state:
        st.session_state['colorscale'] = st.sidebar.selectbox('Color scale', available_colorscales,
                                                              index=available_colorscales.index(default_colorscale))
    else:
        value = st.session_state['colorscale']
        st.session_state['colorscale'] = st.sidebar.selectbox('Color scale', available_colorscales,
                                                              index=available_colorscales.index(value))

    new_user_preferences_df = pd.DataFrame([['project_name', st.session_state['project_name']],
                                            ['path_to_outdirs', st.session_state['path_to_outdirs']],
                                            ['plot_height', st.session_state['plot_height']],
                                            ['plot_width', st.session_state['plot_width']],
                                            ['template', st.session_state['template']],
                                            ['show_legend', st.session_state['show_legend']],
                                            ['font_size', st.session_state['font_size']],
                                            ['clustering_unit', st.session_state['clustering_unit']],
                                            ['scatter_size', st.session_state['scatter_size']],
                                            ['color_1', st.session_state['color_1']],
                                            ['color_2', st.session_state['color_2']],
                                            ['colorsequence', st.session_state['colorsequence']],
                                            ['colorscale', st.session_state['colorscale']],
                                            ], columns=['Variable', 'Value'])

    new_user_preferences_dict = {
        'project_name': st.session_state['project_name'],
        'path_to_outdirs': path_to_projects,
        'plot_height': st.session_state['plot_height'],
        'plot_width': st.session_state['plot_width'],
        'show_legend': st.session_state['show_legend'],
        'template': st.session_state['template'],
        'font_size': st.session_state['font_size'],
        'clustering_unit': st.session_state['clustering_unit'],
        'scatter_size': st.session_state['scatter_size'],
        'color_1': st.session_state['color_1'],
        'color_2': st.session_state['color_2'],
        'colorsequence': st.session_state['colorsequence'],
        'colorscale': st.session_state['colorscale']
    }

    st.sidebar.write(""" ### Save """)

    if st.sidebar.button('Save preferences'):
        updated_user_preferences_df = pd.DataFrame(list(new_user_preferences_dict.items()), columns=['Variable', 'Value'])
        updated_user_preferences_df.to_excel(user_preferences_xlsx, index=False)
        st.sidebar.write('Saved preferences...!')

    if st.sidebar.button('Help'):
        webbrowser.open_new_tab(
            'https://sites.google.com/d/1JHe4k_-j7X0_XHj0eQ9HRJYfLwb-sVt4/p/1mSo5PZ0O5oYSz6jLGLHS5wAKyw0ZBo_C/edit')

    return {
        'projects_dict': projects_dict,
        'cache_df': cache_df,
        'TaXon_table_df': TaXon_table_df,
        'new_user_preferences_dict': new_user_preferences_dict
        }



## variables for testing: IGNORE
path_to_outdirs = Path('/Users/tillmacher/Desktop/TTT_projects/AA_test/')
taxon_table_xlsx = Path('/Users/tillmacher/Desktop/TTT_projects/AA_test/TaXon_tables/GeDNA_MZB_20_21_taxon_table_cons_NCsub_MZB_merged_shared.xlsx')
taxon_table_df = pd.read_excel(taxon_table_xlsx).fillna('')
taxon_table_df = strip_traits(taxon_table_df)
samples = list(taxon_table_df.columns[9:])
metadata_df = pd.read_excel(taxon_table_xlsx, sheet_name='Metadata Table').fillna('')
selected_metadata = 'Venn3'
traits_df = collect_traits(taxon_table_df)
selected_traits = ''
user_settings = {
    'project_name': 'AA_test',
    'path_to_outdirs': path_to_outdirs,
    'plot_height': 800,
    'plot_width': 800,
    'template': 'simple_white',
    'font_size': 16,
    'clustering_unit': 'ESVs',
    'scatter_size': 10,
    'color_1': 'Navy',
    'color_2': 'Teal',
    'colorsequence': 'Plotly',
    'colorscale': 'Blues',
    'show_legend':True,
    'show_xaxis':True,
    'show_yaxis':False,
    }
tool_settings = {'selected_metadata': 'Venn2', 'taxonomic_level': 'Species', 'metric': 'Jaccard', 'dimensions':'2', 'draw_outlines':True, }
# taxon_table_df = simple_taxontable(taxon_table_xlsx, taxon_table_df, samples, metadata_df, False)


selected_axes = ['PC1 (57.86%)', 'PC2 (14.89%)', 'PC3 (13.19%)']

# Use color map based on user settings
colorscale_name = user_settings.get('colorsequence', 'Plotly')
colors = getattr(px.colors.qualitative, colorscale_name, px.colors.qualitative.Plotly)

# Use color map based on user settings
colorscale_name = user_settings.get('colorscale', 'Viridis')

tool_settings['perlodes_TTT_conversion_xlsx'] = "/Library/Frameworks/Python.framework/Versions/3.10/lib/python3.10/site-packages/taxontabletools2/WFD_conversion/perlodes_TTT_conversion.xlsx"
tool_settings['selected_metadata'] = 'Type'
tool_settings['presence_absence'] = True

import plotly.colors

# Sample 100 colors from the colorscale
num_colors = 100
color_list = plotly.colors.sample_colorscale(user_settings['colorscale'], [i / (num_colors - 1) for i in range(num_colors)])



import pandas as pd
import math
import numpy as np
from taxontabletools2.utilities import collect_traits
from taxontabletools2.utilities import strip_traits
from taxontabletools2.utilities import collect_metadata
from taxontabletools2.utilities import load_df
from taxontabletools2.utilities import collect_replicates
from taxontabletools2.utilities import export_taxon_table

def sum_if_less_than_X_zeros(row, cutoff):
    if (row != 0).sum() < cutoff:
        return 0
    else:
        return row.sum()

def replicate_merging(taxon_table_xlsx, taxon_table_df, samples, metadata_df, traits_df, cutoff, replicates_dict):
    ## create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    ## collect clean sample names
    unique_samples, suffixes = replicates_dict
    clean_samples = sorted(set([sample.replace('_' + suffix, '') for sample in samples for suffix in suffixes if suffix in sample]))

    ## merge replicates
    for sample in clean_samples:
        replicates = [f'{sample}_{i}' for i in suffixes]
        if all(value in samples for value in replicates):
            sub_df = taxon_table_df[replicates]
            taxon_table_df[sample] = sub_df.apply(sum_if_less_than_X_zeros, axis=1, args=(cutoff,))
            taxon_table_df.drop(columns=replicates, inplace=True)

    suffix = 'merged'
    export_taxon_table(taxon_table_xlsx, taxon_table_df, traits_df, metadata_df, suffix)

def negative_control_subtraction(taxon_table_xlsx, taxon_table_df, samples, metadata_df, traits_df, mode, NC_list):
    ## create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()


    ## collect samples
    samples_keep = [i for i in samples if i not in NC_list]
    # filter for samples
    sample_df = taxon_table_df[samples_keep]

    ## calculate NC reads
    if mode == 'Sum of all NCs':
        NC_df = taxon_table_df[NC_list].sum(axis=1)
    elif mode == 'Maximum in all NCs':
        NC_df = taxon_table_df[NC_list].max(axis=1)
    elif mode == 'Average in all NCs':
        NC_df = taxon_table_df[NC_list].mean(axis=1)
    result_df = sample_df.subtract(NC_df, axis=0)
    result_df = result_df.clip(lower=0)
    result_df['ID'] = taxon_table_df['ID']

    ## merge tables again
    taxon_table_df.drop(samples, axis=1, inplace=True)
    merged_taxon_table_df = taxon_table_df.merge(result_df, on='ID', how='inner')

    ## export table
    suffix = 'NCsub'
    export_taxon_table(taxon_table_xlsx, merged_taxon_table_df, traits_df, metadata_df, suffix)

def read_based_filter(taxon_table_xlsx, taxon_table_df, samples, metadata_df, traits_df, mode, read_cutoff):
    ## create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()


    ## filter reads
    if mode == 'Absolute':
        taxon_table_df[samples] = taxon_table_df[samples].where(taxon_table_df[samples] >= read_cutoff, 0)
    elif mode == 'Relative':
        for sample in samples:
            sample_cutoff = math.ceil(taxon_table_df[sample].sum() * read_cutoff)
            taxon_table_df[sample] = taxon_table_df[sample].where(taxon_table_df[sample] >= sample_cutoff, 0)

    ## export table
    suffix = 'filtered'
    export_taxon_table(taxon_table_xlsx, taxon_table_df, traits_df, metadata_df, suffix)

def read_based_normalisation(taxon_table_xlsx, taxon_table_df, samples, metadata_df, traits_df, sub_sample_size):
    ## create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    ## count the lowest number of reads to use as threshold
    reads_list = [sum(taxon_table_df[i].values.tolist()) for i in samples]

    ## convert sub_sample_size to int
    sub_sample_size = int(sub_sample_size)

    ## collect all OTUs
    OTU_list = taxon_table_df["ID"].values.tolist()

    df_out = pd.DataFrame()
    required_columns = taxon_table_df.columns.tolist()[0:10]
    df_out[required_columns] = taxon_table_df[required_columns]

    warnings = []

    for sample in samples:
        ## filter sample from data
        read_df = taxon_table_df[[sample, "ID"]]
        ## drop empty OTUs
        read_df = read_df[read_df[sample] != 0]
        ## check if sample can be normalized, otherwise just keep all reads and OTUs
        if sub_sample_size <= sum(read_df[sample].values.tolist()):
            ## create read list to draw the subsamples from
            read_list = pd.Series(np.repeat(read_df['ID'].to_list(), read_df[sample].to_list()))
            ## create a random subsample
            sub_sample = read_list.sample(n = sub_sample_size)
            ## count the number of reads per OTU
            sub_sample_reads = dict(pd.Series(sub_sample).value_counts())
            ## create a sorted OTU list
            OTU_sample_list = []
            for OTU in OTU_list:
                if OTU in sub_sample_reads.keys():
                    OTU_sample_list.append(sub_sample_reads[OTU])
                else:
                    OTU_sample_list.append(0)
        else:
            OTU_sample_list = taxon_table_df[[sample]]
            warnings.append(sample)

        ## add OTUs to dataframe
        df_out[sample] = OTU_sample_list

    ## export table
    suffix = 'norm'
    export_taxon_table(taxon_table_xlsx, df_out, traits_df, metadata_df, suffix)

def taxonomic_filtering(taxon_table_xlsx, taxon_table_df, samples, metadata_df, traits_df, mode, taxa_list, taxonomic_level, suffix):
    ## create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    # keep
    if mode == 'Keep':
        filtered_df = taxon_table_df.loc[taxon_table_df[taxonomic_level].isin(taxa_list)]
    # exclude
    else:
        filtered_df = taxon_table_df.loc[~taxon_table_df[taxonomic_level].isin(taxa_list)]

    ## export table
    export_taxon_table(taxon_table_xlsx, filtered_df, traits_df, metadata_df, suffix)

def sample_filtering(taxon_table_xlsx, taxon_table_df, samples, metadata_df, traits_df, mode, samples_list, suffix):
    ## create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    # keep
    if mode == 'Remove':
        pop_list = [i for i in samples if i in samples_list]
        filtered_df = taxon_table_df.drop(pop_list, axis=1)

    # exclude
    else:
        pop_list = [i for i in samples if i not in samples_list]
        filtered_df = taxon_table_df.drop(pop_list, axis=1)

    ## export table
    export_taxon_table(taxon_table_xlsx, filtered_df, traits_df, metadata_df, suffix)








import numpy as np
import pandas as pd
from pathlib import Path
from taxontabletools2.utilities import simple_taxontable
from taxontabletools2.utilities import filter_taxontable
from taxontabletools2.utilities import taxon_richness_df
import plotly.graph_objects as go
import statsmodels.api as sm

def time_series_richness_with_ci(path_to_outdirs, taxon_table_xlsx, taxon_table_df, samples, metadata_df,
                                 selected_metadata, traits_df,
                                 selected_traits, user_settings, tool_settings):
    ## create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    ## collect tool-specific settings
    selected_metadata = tool_settings['selected_metadata']
    taxonomic_level_1 = tool_settings['taxonomic_level_1']
    taxonomic_level_2 = tool_settings['taxonomic_level_2']
    n_bootstraps = tool_settings['n_bootstraps']
    ci_level = tool_settings['ci_level']

    richness_df = taxon_richness_df(taxon_table_df, samples, taxonomic_level_1)

    ## collect the number of categories
    samples_sorted = metadata_df['Sample'].values.tolist()

    # Prepare x-values and y-values (using actual sample names as x-values)
    x_values = samples_sorted  # Use the sample names instead of index numbers
    y_values = richness_df[samples_sorted].values.tolist()[0]

    # Plot original data points
    fig = go.Figure()
    fig.add_trace(go.Scatter(x=x_values, y=y_values, name='Original Data', mode='markers', marker_color=user_settings['color_1'], marker=dict(size=user_settings['scatter_size']) ))

    # Calculate LOESS (Locally Estimated Scatterplot Smoothing) for the original data
    # Since x_values are now sample names, we need numerical indices for LOESS smoothing
    x_numeric = list(range(len(x_values)))
    loess_smoothed = sm.nonparametric.lowess(endog=y_values, exog=x_numeric, frac=0.3)
    loess_x_numeric = loess_smoothed[:, 0]
    loess_y = loess_smoothed[:, 1]

    # Map numerical x-values back to sample names
    loess_x = [x_values[int(i)] for i in loess_x_numeric]

    # Add LOESS-smoothed line to the plot
    fig.add_trace(go.Scatter(x=loess_x, y=loess_y, name='LOESS Smoothing', mode='lines', marker_color=user_settings['color_2'], marker=dict(size=user_settings['scatter_size']) ))

    # Bootstrapping to calculate confidence intervals
    bootstrapped_fits = []

    for _ in range(n_bootstraps):
        # Resample the data with replacement
        resample_indices = np.random.choice(range(len(x_numeric)), size=len(x_numeric), replace=True)
        resample_x_numeric = np.array(x_numeric)[resample_indices]
        resample_y = np.array(y_values)[resample_indices]

        # Fit LOESS on resampled data
        loess_bootstrap = sm.nonparametric.lowess(endog=resample_y, exog=resample_x_numeric, frac=0.3)

        # Interpolate to original x-numeric values
        interp_bootstrap_y = np.interp(x_numeric, loess_bootstrap[:, 0], loess_bootstrap[:, 1])
        bootstrapped_fits.append(interp_bootstrap_y)

    # Convert list to numpy array for easier calculations
    bootstrapped_fits = np.array(bootstrapped_fits)

    # Calculate the confidence intervals (CI) at each x-value
    lower_bound = np.percentile(bootstrapped_fits, (100 - ci_level) / 2, axis=0)
    upper_bound = np.percentile(bootstrapped_fits, 100 - (100 - ci_level) / 2, axis=0)

    # Add shaded region for the confidence interval
    fig.add_trace(go.Scatter(
        x=x_values + x_values[::-1],  # x values forward and backward for the fill area
        y=np.concatenate([upper_bound, lower_bound[::-1]]),  # upper bound followed by reversed lower bound
        fill='toself',
        fillcolor='rgba(0,100,80,0.2)',  # Semi-transparent color
        line=dict(color='rgba(255,255,255,0)'),
        showlegend=False,
        name=f'{ci_level}% Confidence Interval'
    ))

    # Update layout
    fig.update_xaxes(dtick='linear')
    fig.update_yaxes(rangemode='tozero')
    fig.update_layout(
        title=f'Richness ({taxonomic_level_1}) with LOESS Smoothing and {ci_level}% Confidence Interval',
        template=user_settings['template'],
        font_size=user_settings['font_size'],
        yaxis_title=f'Richness ({taxonomic_level_1})',
        yaxis_title_font=dict(size=user_settings['font_size']),
        yaxis_tickfont=dict(size=user_settings['font_size']),
        xaxis_title='Sample',
        xaxis_tickvals=x_values,  # Ensure sample names are shown as tick labels
        xaxis_ticktext=x_values,  # Use sample names as tick labels
        xaxis_tickfont=dict(size=user_settings['font_size'])
    )

    # if user_settings['show_yaxis'] == False:
    #    fig.update_xaxes(showticklabels=False)

    return fig






import pandas as pd
import math
import numpy as np
from pathlib import Path
import streamlit as st
from taxontabletools2.utilities import collect_traits
from taxontabletools2.utilities import strip_traits
from taxontabletools2.utilities import collect_metadata
from taxontabletools2.utilities import load_df
from taxontabletools2.utilities import collect_replicates
from taxontabletools2.utilities import export_taxon_table
from taxontabletools2.utilities import simple_taxontable
from taxontabletools2.utilities import update_taxon_table

def presence_absence_table(taxon_table_xlsx, taxon_table_df, samples, metadata_df, traits_df):
    ## create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    ## set all element to 0/1
    taxon_table_df[samples] = taxon_table_df[samples].apply(lambda x: (x != 0).astype(int))

    ## export table
    suffix = 'PA'
    export_taxon_table(taxon_table_xlsx, taxon_table_df, traits_df, metadata_df, suffix)

def simplify_table(taxon_table_xlsx, taxon_table_df, samples, metadata_df, traits_df):
    # merge duplicate OTUs
    save = False
    filtered_df = simple_taxontable(taxon_table_xlsx, taxon_table_df, samples, metadata_df, save)
    ## export the dataframe
    new_TaXon_table_xlsx = str(Path(taxon_table_xlsx)).replace('.xlsx', '_simplified.xlsx')
    with pd.ExcelWriter(new_TaXon_table_xlsx) as writer:
        filtered_df.to_excel(writer, sheet_name='Taxon Table', index=False)
        metadata_df.to_excel(writer, sheet_name='Metadata Table', index=False)

def add_traits_from_file(taxon_table_xlsx, taxon_table_df, samples, metadata_df, traits_df, new_traits_df, taxon_col):
    ## create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    for row in new_traits_df.values.tolist():
        taxon = row[0]
        traits = row[1:]
        for trait, trait_name in zip(traits, new_traits_df.columns.tolist()[1:]):
            # Find the index of the rows in taxon_table_df that match the taxon
            indices = taxon_table_df[taxon_table_df[taxon_col] == taxon].index
            # Add traits to each matching row in taxon_table_df
            for index in indices:
                traits_df.loc[index, trait_name] = trait

    traits_df = traits_df.fillna('')
    update_taxon_table(taxon_table_xlsx, taxon_table_df, traits_df, metadata_df, '')

def sort_samples(taxon_table_xlsx, taxon_table_df, samples, metadata_df, traits_df):
    ## create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    # Collect the sorted samples
    sorted_samples = metadata_df['Sample'].values.tolist()
    taxon_table_df_sorted = taxon_table_df[taxon_table_df.columns.tolist()[:9] + sorted_samples]

    update_taxon_table(taxon_table_xlsx, taxon_table_df_sorted, traits_df, metadata_df, '')

def rename_samples(taxon_table_xlsx, taxon_table_df, samples, metadata_df, traits_df, selected_metadata):
    # Create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    # Collect the new sample names from the selected_metadata column
    sorted_samples = metadata_df['Sample'].values.tolist()
    new_names = metadata_df[selected_metadata].values.tolist()

    # Rename columns in taxon_table_df using the old and new sample names
    for old, new in zip(sorted_samples, new_names):
        taxon_table_df.rename(columns={old: new}, inplace=True)

    # Update the metadata DataFrame with the old and new sample names
    metadata_df['Old names'] = sorted_samples
    metadata_df['Sample'] = new_names

    # Call the update function to apply changes
    update_taxon_table(taxon_table_xlsx, taxon_table_df, traits_df, metadata_df, '')




import streamlit as st
import pandas as pd
import numpy as np
import scipy, subprocess, webbrowser
from pathlib import Path
import glob, sys, time, statistics, os.path
import seqconverter


########################################################################################################################
def load_df(table):
    df = pd.read_excel(table).fillna('')
    return df

def reduce_taxontable(df, taxonomic_level, samples, meta_data_to_test):

    if meta_data_to_test == '':
        ## extract the relevant data
        df = df[[taxonomic_level] + samples]
        ## define an aggregation function to combine multiple hit of one taxonimic level
        aggregation_functions = {}
        ## define samples functions
        for sample in samples:
            ## 'sum' will calculate the sum of p/a data
            aggregation_functions[sample] = 'sum'
        ## define taxon level function
        aggregation_functions[taxonomic_level] = 'first'
        ## create condensed dataframe
        df = df.groupby(df[taxonomic_level]).aggregate(aggregation_functions)
        if '' in df.index:
            df = df.drop('')

        return df

    else:
        ## collect metadata for each taxon
        taxon_metadata_dict = {}
        for row in df[[taxonomic_level, meta_data_to_test]].values.tolist():
            taxon = row[0]
            metadata = row[1]
            taxon_metadata_dict[taxon] = metadata

        ## extract the relevant data
        df = df[[taxonomic_level] + samples]
        ## define an aggregation function to combine multiple hit of one taxonimic level
        aggregation_functions = {}
        ## define samples functions
        for sample in samples:
            ## 'sum' will calculate the sum of p/a data
            aggregation_functions[sample] = 'sum'
        ## define taxon level function
        aggregation_functions[taxonomic_level] = 'first'
        ## create condensed dataframe
        df = df.groupby(df[taxonomic_level]).aggregate(aggregation_functions)
        if '' in df.index:
            df = df.drop('')

        ## add metadata back to taxa
        metadata_col = []
        for taxon in df[taxonomic_level].values.tolist():
            metadata_col.append(taxon_metadata_dict[taxon])
        df[meta_data_to_test] = metadata_col

        return df

def strip_traits(df):
    " Strip the additional metadata from the dataframe to produce a clean TaxonTable "

    ## position of seq == acts as separator
    seq_loc = df.columns.get_loc("Seq")

    ## samples are always assending the seq column
    samples = df.columns.tolist()[seq_loc+1:]

    ## standard columns
    standard_columns = ['ID', 'Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species', 'Similarity', 'Seq']

    ## extract legacy TTT frame
    df = df[standard_columns + samples]

    return df

def collect_traits(df):
    " Collect the additional metadata from the dataframe to be able to place it back later "

    ## position of status and seq == acts as separator
    seq_loc = df.columns.get_loc("Seq")
    status = df.columns.get_loc("Similarity")

    ## metadata is always located between status and seq
    metadata = df.columns.tolist()[status+1:seq_loc]

    ## collect IDs and metadata
    df = df[['ID'] + metadata]

    return df

def add_traits(df, traits_df):
    " Collect the additional metadata from the dataframe to add them back to a df "

    for metadata in  traits_df.columns.tolist()[1:]:
        status_loc = df.columns.get_loc("Seq")
        ## loope through IDs and find all that still remain in the altered Taxon table (e.g. after filtering less OTUs might be present)
        sub_df = pd.DataFrame([i for i in traits_df[['ID', metadata]].values.tolist() if i[0] in df['ID'].values.tolist()], columns = ['ID', metadata])
        ## insert sub_df containing the respective metadata after the Status column
        df.insert(status_loc, str(metadata), sub_df[metadata].values.tolist(), True)

    return df

def collect_metadata(taxon_table_xlsx):
    try:
        metadata_df = pd.read_excel(taxon_table_xlsx, sheet_name='Metadata table')
    except ValueError:
        metadata_df = pd.DataFrame(columns=['Sample', 'Metadata'])
    return metadata_df

def export_taxon_table(taxon_table_xlsx, taxon_table_df, traits_df, metadata_df, suffix):
    ## collect remaining samples and filter metadata_df
    seq_loc = taxon_table_df.columns.tolist().index('Seq')
    samples = taxon_table_df.columns.tolist()[seq_loc+1::]
    filtered_metadata_df = metadata_df.loc[metadata_df['Sample'].isin(samples)]

    ## merge taxon table and trait table
    merged_taxon_table_df = taxon_table_df.merge(traits_df, on='ID', how='inner')
    cols = list(merged_taxon_table_df.columns)
    seq_index = cols.index('Seq')
    traits_cols = traits_df.columns.tolist()
    traits_cols.remove('ID')  # remove 'ID' as it is already in `taxon_table_df`
    [cols.remove(trait) for trait in traits_cols]
    new_cols = cols[:seq_index] + traits_cols + cols[seq_index:]
    merged_taxon_table_df = merged_taxon_table_df[new_cols]

    ## remove empty OTUs from dataframe
    mask = (merged_taxon_table_df[samples] != 0).any(axis=1)
    merged_taxon_table_df = merged_taxon_table_df[mask]

    ## create a new file name
    new_TaXon_table_xlsx = Path(str(Path(taxon_table_xlsx)).replace('.xlsx', f'_{suffix}.xlsx'))

    ## export the dataframe
    with pd.ExcelWriter(new_TaXon_table_xlsx) as writer:
        merged_taxon_table_df.to_excel(writer, sheet_name='Taxon Table', index=False)
        filtered_metadata_df.to_excel(writer, sheet_name='Metadata Table', index=False)

def update_taxon_table(taxon_table_xlsx, taxon_table_df, traits_df, metadata_df, suffix):
    ## collect remaining samples and filter metadata_df
    seq_loc = taxon_table_df.columns.tolist().index('Seq')
    samples = taxon_table_df.columns.tolist()[seq_loc+1::]
    filtered_metadata_df = metadata_df.loc[metadata_df['Sample'].isin(samples)]

    ## merge taxon table and trait table
    merged_taxon_table_df = taxon_table_df.merge(traits_df, on='ID', how='inner')
    cols = list(merged_taxon_table_df.columns)
    seq_index = cols.index('Seq')
    traits_cols = traits_df.columns.tolist()
    traits_cols.remove('ID')  # remove 'ID' as it is already in `taxon_table_df`
    [cols.remove(trait) for trait in traits_cols]
    new_cols = cols[:seq_index] + traits_cols + cols[seq_index:]
    merged_taxon_table_df = merged_taxon_table_df[new_cols]

    ## remove empty OTUs from dataframe
    mask = (merged_taxon_table_df[samples] != 0).any(axis=1)
    merged_taxon_table_df = merged_taxon_table_df[mask]

    ## export the dataframe
    with pd.ExcelWriter(taxon_table_xlsx) as writer:
        merged_taxon_table_df.to_excel(writer, sheet_name='Taxon Table', index=False)
        filtered_metadata_df.to_excel(writer, sheet_name='Metadata Table', index=False)

def simple_taxontable(taxon_table_xlsx, taxon_table_df, samples, metadata_df, save):
    ## create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()

    ## add unique taxa
    taxon_table_df['Taxon'] = [' '.join(i) for i in taxon_table_df[['Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species']].values.tolist()]
    unique_taxa = taxon_table_df['Taxon'].drop_duplicates().values.tolist()
    new_df_list = []
    n_OTUs_list = []
    for i, taxon in enumerate(unique_taxa):
        sub_df = taxon_table_df.loc[taxon_table_df['Taxon'] == taxon].copy()
        sub_df.pop('Taxon')
        sub_df.loc[:, 'sub_df_sort'] = sub_df[samples].sum(axis=1)
        sub_df = sub_df.sort_values('sub_df_sort', ascending=False)
        sub_df.pop('sub_df_sort')
        n_OTUs_list.append(len(sub_df))
        samples = sub_df.columns.tolist()[sub_df.columns.tolist().index('Seq')+1:]
        reads = sub_df[samples].sum(axis=0)
        OTU_name = f'spOTU_{i}'
        new_row = [OTU_name] + sub_df.iloc[0].values.tolist()[1:sub_df.columns.tolist().index('Seq')+1] + reads.values.tolist()
        new_df_list.append(new_row)
    taxon_table_df.pop('Taxon')
    new_df = pd.DataFrame(new_df_list, columns=taxon_table_df.columns.tolist())
    new_df['Similarity'] = 100
    new_df.insert(9, "Merged", n_OTUs_list)
    new_df['Seq'] = 'Please refer to the non-simplified table!'

    ## create a new file name
    new_TaXon_table_xlsx = str(Path(taxon_table_xlsx)).replace('.xlsx', '_simplified.xlsx')

    if save == True:
        ## export the dataframe
        with pd.ExcelWriter(new_TaXon_table_xlsx) as writer:
            new_df.to_excel(writer, sheet_name='Taxon Table', index=False)
            metadata_df.to_excel(writer, sheet_name='Metadata Table', index=False)
    else:
        return new_df

def filter_taxontable(taxon_table_df, sample_subset, taxonomic_level):
    sub_df = taxon_table_df[[taxonomic_level] + sample_subset]
    sub_df = sub_df.loc[sub_df[taxonomic_level] != '']
    merged_df = sub_df.groupby(taxonomic_level, as_index=False).sum()
    filtered_df = merged_df[(merged_df.iloc[:, 1:].sum(axis=1) != 0)]
    return  filtered_df

def filter_taxontable_2(taxon_table_df, sample_subset, taxonomic_level_1, taxonomic_level_2):
    sub_df = taxon_table_df[[taxonomic_level_1, taxonomic_level_2] + sample_subset]
    sub_df = sub_df[(sub_df[taxonomic_level_1] != '') | (sub_df[taxonomic_level_2] != '')]
    sub_df['Combined_Taxon'] = sub_df[taxonomic_level_1] + ' ' + sub_df[taxonomic_level_2]
    merged_df = sub_df.groupby(['Combined_Taxon', taxonomic_level_1, taxonomic_level_2], as_index=False).sum()
    filtered_df = merged_df[(merged_df.iloc[:, 3:].sum(axis=1) != 0)]
    filtered_df.drop(columns=['Combined_Taxon'], inplace=True)
    column_order = [taxonomic_level_1, taxonomic_level_2] + sample_subset
    filtered_df = filtered_df[column_order]
    return filtered_df

########################################################################################################################

def taxon_richness_df(taxon_table_df, samples, taxonomic_level_1):
    richness_dict = {}  # Number of taxa per sample
    for sample in samples:
        sub_df = taxon_table_df[[sample, taxonomic_level_1]]
        sub_df = sub_df.loc[sub_df[sample] != 0]
        richness_dict[sample] = [len(sub_df[taxonomic_level_1].drop_duplicates())]

    richness_df = pd.DataFrame(richness_dict, index=[taxonomic_level_1])
    return richness_df

def taxon_reads_df(taxon_table_df, samples, taxonomic_level_1):
    reads_dict = {}  # Number of reads per taxon per sample
    all_taxa = sorted([i for i in taxon_table_df[taxonomic_level_1].drop_duplicates().values.tolist() if i != ''])

    for sample in samples:
        sub_df = taxon_table_df[[sample, taxonomic_level_1]]
        sub_df = sub_df.loc[sub_df[sample] != 0]

        reads_list = []
        for taxon in all_taxa:
            taxon_df = sub_df.loc[sub_df[taxonomic_level_1] == taxon]
            reads = taxon_df[sample].sum()
            reads_list.append(reads)

        reads_dict[sample] = reads_list

    reads_df = pd.DataFrame(reads_dict, index=all_taxa)
    return reads_df

def taxon_read_proportions_df(taxon_table_df, samples, taxonomic_level_1):
    read_proportions_dict = {}  # Read proportion per taxon per sample
    all_taxa = sorted([i for i in taxon_table_df[taxonomic_level_1].drop_duplicates().values.tolist() if i != ''])

    for sample in samples:
        sub_df = taxon_table_df[[sample, taxonomic_level_1]]
        sub_df = sub_df.loc[sub_df[sample] != 0]
        sample_reads = sub_df[sample].sum()

        read_proportions_list = []
        for taxon in all_taxa:
            taxon_df = sub_df.loc[sub_df[taxonomic_level_1] == taxon]
            reads = taxon_df[sample].sum()
            read_proportions_list.append(reads / sample_reads * 100 if sample_reads != 0 else 0)

        read_proportions_dict[sample] = read_proportions_list

    read_proportions_df = pd.DataFrame(read_proportions_dict, index=all_taxa)
    return read_proportions_df

def taxon_occurrence_df(taxon_table_df, samples, taxonomic_level_1):
    occurrence_dict = {}  # Occurrence (1/0) per taxon per sample
    all_taxa = sorted([i for i in taxon_table_df[taxonomic_level_1].drop_duplicates().values.tolist() if i != ''])

    for sample in samples:
        sub_df = taxon_table_df[[sample, taxonomic_level_1]]
        sub_df = sub_df.loc[sub_df[sample] != 0]

        occurrence_list = []
        for taxon in all_taxa:
            taxon_df = sub_df.loc[sub_df[taxonomic_level_1] == taxon]
            reads = taxon_df[sample].sum()
            occurrence_list.append(1 if reads != 0 else 0)

        occurrence_dict[sample] = occurrence_list

    occurrence_df = pd.DataFrame(occurrence_dict, index=all_taxa)
    return occurrence_df

def taxon_richness_per_higher_taxon_df(taxon_table_df, samples, taxonomic_level_1, taxonomic_level_2):
    richness_per_taxon_dict = {}  # Richness per higher taxon (e.g., family)
    all_taxa = sorted([i for i in taxon_table_df[taxonomic_level_2].drop_duplicates().values.tolist() if i != ''])

    for sample in samples:
        sub_df = taxon_table_df[[sample, taxonomic_level_1, taxonomic_level_2]]
        sub_df = sub_df.loc[sub_df[sample] != 0]

        richness_per_taxon_list = []
        for taxon in all_taxa:
            taxon_df = sub_df.loc[sub_df[taxonomic_level_2] == taxon]
            richness_per_taxon_list.append(len(taxon_df[taxonomic_level_1].drop_duplicates()))

        richness_per_taxon_dict[sample] = richness_per_taxon_list

    richness_per_taxon_df = pd.DataFrame(richness_per_taxon_dict, index=all_taxa)
    return richness_per_taxon_df if taxonomic_level_2 else None

########################################################################################################################
def collect_replicates(samples):

    unique_names = sorted(list(set(['_'.join(i.split('_')[:-1]) for i in samples])))
    suffixes = sorted(list(set([i.split('_')[-1] for i in samples])))

    return [unique_names, suffixes]

def check_replicates(replicates_dict, samples):
    unique_samples, replicate_suffixes = replicates_dict
    expected_samples = [sample+'_'+suffix for sample in unique_samples for suffix in replicate_suffixes]
    missing_samples = set(expected_samples) - set(samples)
    test = sorted(samples) == sorted(expected_samples)
    return test, missing_samples

########################################################################################################################
def export_plot(path_to_outdirs, folder, file_name, suffix, fig, settings):
    pass

########################################################################################################################
def spearman(a, b):
    # spearman's rho
    results = scipy.stats.spearmanr(a, b)
    spearman_p = results[1]
    if spearman_p <= 0.05:
        spearman_rho = str(round(results[0], 3)) + "*"
    else:
        spearman_rho = str(round(results[0], 3))
    return spearman_rho

########################################################################################################################
def glob_available_taxon_tables(path_to_outdirs):
    TaXon_tables_dict = {}
    tables = sorted(glob.glob(str(Path(path_to_outdirs).joinpath('TaXon_tables', '*.xlsx'))), key=os.path.getmtime)[::-1]
    for table in tables:
        TaXon_tables_dict[Path(table).stem] = Path(table)
    return TaXon_tables_dict

########################################################################################################################
def collect_sample_stats(TaXon_table_df, TaXon_table_name, traits_df, samples, metadata_df):
    ## information about the table
    n_samples = len(samples)
    n_OTUs = len(TaXon_table_df['ID'])
    n_phyla = len(set([i for i in TaXon_table_df['Phylum'].values.tolist() if i != '']))
    n_classes = len(set([i for i in TaXon_table_df['Class'].values.tolist() if i != '']))
    n_orders = len(set([i for i in TaXon_table_df['Order'].values.tolist() if i != '']))
    n_families = len(set([i for i in TaXon_table_df['Family'].values.tolist() if i != '']))
    n_genera = len(set([i for i in TaXon_table_df['Genus'].values.tolist() if i != '']))
    n_species = len(set([i for i in TaXon_table_df['Species'].values.tolist() if i != '']))
    n_traits = len(traits_df.columns[1:])
    n_reads = sum([sum(i[10:]) for i in TaXon_table_df.values.tolist()])
    cols = ['Samples', 'ID', 'Phyla', 'Classes', 'Orders', 'Families', 'Genera', 'Species', 'Traits',
            'Reads']
    row = [[n_samples, n_OTUs, n_phyla, n_classes, n_orders, n_families, n_genera, n_species, n_traits, n_reads]]
    stats_df = pd.DataFrame(row, columns=cols, index=['#'])

    # collect information about each sample in the table
    samples_info_list = []
    for sample in samples:
        tmp = pd.DataFrame([i for i in TaXon_table_df[
            ['ID', 'Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species', sample]].values.tolist() if i[-1] != 0],
                           columns=['ID', 'Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species', sample])
        n_OTUs = len(tmp['ID'])
        n_phyla = len(set([i for i in tmp['Phylum'].values.tolist() if i != '']))
        n_classes = len(set([i for i in tmp['Class'].values.tolist() if i != '']))
        n_orders = len(set([i for i in tmp['Order'].values.tolist() if i != '']))
        n_families = len(set([i for i in tmp['Family'].values.tolist() if i != '']))
        n_genera = len(set([i for i in tmp['Genus'].values.tolist() if i != '']))
        n_species = len(set([i for i in tmp['Species'].values.tolist() if i != '']))
        n_reads = sum(tmp[sample])
        samples_info_list.append(
            [sample, n_OTUs, n_phyla, n_classes, n_orders, n_families, n_genera, n_species, n_reads])

    sample_stats_df = pd.DataFrame(samples_info_list,
                                   columns=['Samples', 'ID', 'Phyla', 'Classes', 'Orders',
                                            'Families', 'Genera', 'Species', 'Reads'])

    # Calculate the mean for each numeric column
    averages = sample_stats_df.select_dtypes(include=[np.number]).mean()
    # Append the averages to the DataFrame using pandas.concat
    sample_stats_df = pd.concat([pd.DataFrame(averages).T, sample_stats_df], ignore_index=True)
    # Set the 'Samples' column of the first row to 'Average'
    sample_stats_df.loc[0, 'Samples'] = 'Average'
    # Replace the index with the 'Samples' column
    sample_stats_df.set_index('Samples', inplace=True)

    # Function to format numbers
    def format_numbers(row):
        if row.name == 'Average':
            return row.round(2).astype(str)
        else:
            return row.astype(int).astype(str)

    # Apply the function to each row
    sample_stats_df = sample_stats_df.apply(format_numbers, axis=1)

    return sample_stats_df

def taxon_table_fasta(taxon_table_xlsx, taxon_table_df):
    taxon_table_xlsx = '/Users/tillmacher/Desktop/TTT_projects/Projects/AA_test/TaXon_tables/All_ESVs_ID_merged_ID_merge_ID_merge_ID_merge_norm.xlsx'
    taxon_table_df = pd.read_excel('/Users/tillmacher/Desktop/TTT_projects/Projects/AA_test/TaXon_tables/All_ESVs_ID_merged_ID_merge_ID_merge_ID_merge_norm.xlsx').fillna('')
    taxon_table_fasta = taxon_table_xlsx.replace('.xlsx', '.fasta')
    f = open(taxon_table_fasta, 'w')
    for row in taxon_table_df[['ID', 'Seq']].values.tolist():
        f.write(f'>{row[0]}\n')
        f.write(f'{row[1]}\n')
    f.close()

########################################################################################################################











import pandas as pd
import numpy as np
from pathlib import Path
from taxontabletools2.utilities import add_traits, filter_taxontable, export_taxon_table
import plotly.graph_objects as go
import plotly.express as px
from scipy.spatial.distance import pdist, squareform
from scipy.spatial import ConvexHull
import streamlit as st
import plotly.colors

def convert_to_perlodes(path_to_outdirs, taxon_table_xlsx, taxon_table_df, samples, metadata_df, selected_metadata,
                        traits_df, selected_traits, user_settings, tool_settings):

    # Make copies of input dataframes to prevent altering the originals
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    # Extract relevant settings from the tool settings
    selected_metadata = tool_settings['selected_metadata']
    presence_absence = tool_settings['presence_absence']

    # Load the operational taxon list for conversion
    perlodes_TTT_conversion_xlsx = st.session_state['perlodes_TTT_conversion_xlsx']
    operational_taxon_list_df = pd.read_excel(Path(perlodes_TTT_conversion_xlsx), sheet_name="TTT import").fillna('')

    # Create a dictionary for freshwater types from metadata
    types_dict = {i[0]: i[1] for i in metadata_df[['Sample', selected_metadata]].values.tolist() if i[1] != ''}
    taxon_table_taxonomy = taxon_table_df.columns.tolist()[0:7]  # Extract taxonomy columns

    # Lists to store hits and dropped OTUs
    hit_list, dropped_list = [], []

    # Loop through the taxon table to match taxonomy with operational taxon list
    for taxonomy in taxon_table_df[taxon_table_taxonomy].values.tolist():
        # Initialize result as empty
        res = taxonomy + ['', '']

        # Check against species, genus, family, order, class, and phylum
        if taxonomy[6] != '' and taxonomy[6] in operational_taxon_list_df['Species'].values.tolist():
            res = taxonomy + operational_taxon_list_df[operational_taxon_list_df['Species'].str.contains(taxonomy[6])][
                ['ID_ART', 'Perlodes']].values.tolist()[0]
        elif taxonomy[5] != '' and taxonomy[5] in operational_taxon_list_df['Genus'].values.tolist():
            res = taxonomy + operational_taxon_list_df[operational_taxon_list_df['Genus'].str.contains(taxonomy[5])][
                ['ID_ART', 'Perlodes']].values.tolist()[0]
        elif taxonomy[4] != '' and taxonomy[4] in operational_taxon_list_df['Family'].values.tolist():
            res = taxonomy + operational_taxon_list_df[operational_taxon_list_df['Family'].str.contains(taxonomy[4])][
                ['ID_ART', 'Perlodes']].values.tolist()[0]
        elif taxonomy[3] != '' and taxonomy[3] in operational_taxon_list_df['Order'].values.tolist():
            res = taxonomy + operational_taxon_list_df[operational_taxon_list_df['Order'].str.contains(taxonomy[3])][
                ['ID_ART', 'Perlodes']].values.tolist()[0]
        elif taxonomy[2] != '' and taxonomy[2] in operational_taxon_list_df['Class'].values.tolist():
            res = taxonomy + operational_taxon_list_df[operational_taxon_list_df['Class'].str.contains(taxonomy[2])][
                ['ID_ART', 'Perlodes']].values.tolist()[0]
        elif taxonomy[1] != '' and taxonomy[1] in operational_taxon_list_df['Phylum'].values.tolist():
            res = taxonomy + operational_taxon_list_df[operational_taxon_list_df['Phylum'].str.contains(taxonomy[1])][
                ['ID_ART', 'Perlodes']].values.tolist()[0]
        else:
            dropped_list.append(taxonomy)  # Append to dropped list if no match is found

        hit_list.append(res)  # Append results for matched taxonomy

    # Create a new DataFrame from the hit list
    hit_df = pd.DataFrame(hit_list, columns=taxon_table_df.columns[0:7].values.tolist() + ['ID_ART', 'TAXON_NAME'])

    # Extract the sample read counts from the taxon table
    samples_reads_df = taxon_table_df[['ID'] + samples]

    # Merge the two DataFrames based on the 'ID' column
    merged_df = pd.merge(hit_df, samples_reads_df, on='ID', how='inner')

    # Extract unique taxa with 'ID_ART' and 'TAXON_NAME'
    unique_taxa_df = merged_df[['ID_ART', 'TAXON_NAME']].drop_duplicates()

    # Initialize the list for storing presence/absence or read sum data
    perlodes_results = []

    # Loop over the unique taxa to calculate read sums or presence/absence
    for taxon_entry in unique_taxa_df.values.tolist():
        taxon_id = taxon_entry[0]
        taxon_name = taxon_entry[1]
        if taxon_name:
            # Calculate the sum of reads for the current taxon across all samples
            read_sums = merged_df[merged_df['TAXON_NAME'].str.contains(taxon_name, regex=False)][samples].sum(
                axis=0).values.tolist()
            # Convert read counts to presence/absence if required
            if presence_absence:
                read_sums = [0 if count == 0 else 1 for count in read_sums]
            # Append the results to the list
            perlodes_results.append([taxon_id, taxon_name] + read_sums)

    # Create header rows for the input file
    # Create a DataFrame for Perlodes output
    perlodes_df = pd.DataFrame(perlodes_results, dtype=object)  # Set dtype to object to avoid type mismatch
    perlodes_df.columns = ["ID_ART", "TAXON_NAME"] + samples

    # Insert header rows for the Perlodes output
    perlodes_df.loc[0] = ["Gewässertyp", ""] + [types_dict[i] for i in samples]  # Water type header
    perlodes_df.loc[1] = ["Taxaliste", ""] + ["original"] * len(samples)  # Taxon list header
    perlodes_df.loc[2] = ["Nutzung", ""] + ["keine"] * len(samples)  # Usage header (none)

    # Update taxon table
    taxon_table_df_perlodes = taxon_table_df.copy()
    traits_df_perlodes = traits_df.copy()
    traits_df_perlodes = pd.merge(hit_df[['ID', 'ID_ART', 'TAXON_NAME']], traits_df_perlodes, on='ID', how='inner')
    export_taxon_table(taxon_table_xlsx, taxon_table_df_perlodes, traits_df_perlodes, metadata_df, 'perlodes')

    # Export Perlodes table
    perlodes_input_xlsx = path_to_outdirs.joinpath('Perlodes', taxon_table_xlsx.stem + '_perlodes.xlsx')
    perlodes_df.to_excel(perlodes_input_xlsx, index=False)
    st.success(f'Wrote Perlodes upload file to: {perlodes_input_xlsx}')




import pandas as pd
import numpy as np
from pathlib import Path
from taxontabletools2.utilities import simple_taxontable
from taxontabletools2.utilities import filter_taxontable
import matplotlib.pyplot as plt
from matplotlib_venn import venn2, venn3
import streamlit as st
import matplotlib.pyplot as plt
from matplotlib_venn import venn2, venn3
from io import BytesIO

def venn(path_to_outdirs, taxon_table_xlsx, taxon_table_df, samples, metadata_df, selected_metadata, traits_df,
         selected_traits, users_settings, tool_settings):
    ## create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    ## collect tool-specific settings
    selected_metadata = tool_settings['selected_metadata']
    taxonomic_level = tool_settings['taxonomic_level']

    ## collect the number of categories
    metadata_df = metadata_df[['Sample', selected_metadata]]
    categories = [i for i in metadata_df[selected_metadata].drop_duplicates().values.tolist() if i != '']
    n_categories = len(categories)

    if n_categories == 2:
        samples_a = metadata_df.loc[metadata_df[selected_metadata] == categories[0]]['Sample'].values.tolist()
        df_a = filter_taxontable(taxon_table_df, samples_a, taxonomic_level)
        species_a = set(df_a[taxonomic_level].values.tolist())
        samples_b = metadata_df.loc[metadata_df[selected_metadata] == categories[1]]['Sample'].values.tolist()
        df_b = filter_taxontable(taxon_table_df, samples_b, taxonomic_level)
        species_b = set(df_b[taxonomic_level].values.tolist())

        a_only = species_a - species_b
        n_a_only = len(a_only)
        shared = species_a & species_b
        n_shared = len(shared)
        b_only = species_b - species_a
        n_b_only = len(b_only)

        # Create the Venn diagram
        plt.figure(figsize=(8, 8))
        venn = venn2(subsets=(n_a_only, n_shared, n_b_only), set_labels=(categories[0], categories[1]))

        # Adjust the font size of set labels
        venn.get_label_by_id('A').set_fontsize(12)  # Font size for "Category A"
        venn.get_label_by_id('B').set_fontsize(12)  # Font size for "Category B"
        # Adjust the font size of subset labels (the numbers inside the circles)
        for text in venn.subset_labels:
            if text:  # Check if the label exists (some subsets might be None)
                text.set_fontsize(users_settings['font_size'])

        # Display the plot in Streamlit
        st.pyplot(plt.gcf())
        plt.close()

    elif n_categories == 3:
        samples_a = metadata_df.loc[metadata_df[selected_metadata] == categories[0]]['Sample'].values.tolist()
        df_a = filter_taxontable(taxon_table_df, samples_a, taxonomic_level)
        species_a = set(df_a[taxonomic_level].values.tolist())

        samples_b = metadata_df.loc[metadata_df[selected_metadata] == categories[1]]['Sample'].values.tolist()
        df_b = filter_taxontable(taxon_table_df, samples_b, taxonomic_level)
        species_b = set(df_b[taxonomic_level].values.tolist())

        samples_c = metadata_df.loc[metadata_df[selected_metadata] == categories[2]]['Sample'].values.tolist()
        df_c = filter_taxontable(taxon_table_df, samples_c, taxonomic_level)
        species_c = set(df_c[taxonomic_level].values.tolist())

        # Calculate the different subsets
        a_only = species_a - (species_b | species_c)
        b_only = species_b - (species_a | species_c)
        c_only = species_c - (species_a | species_b)
        ab_shared = (species_a & species_b) - species_c
        ac_shared = (species_a & species_c) - species_b
        bc_shared = (species_b & species_c) - species_a
        abc_shared = species_a & species_b & species_c

        # Create the Venn diagram
        plt.figure(figsize=(8, 8))
        venn = venn3(subsets=(
        len(a_only), len(b_only), len(ab_shared), len(c_only), len(ac_shared), len(bc_shared), len(abc_shared)),
                     set_labels=(categories[0], categories[1], categories[2]))

        # Adjust the font size of set labels
        venn.get_label_by_id('A').set_fontsize(12)  # Font size for "Category A"
        venn.get_label_by_id('B').set_fontsize(12)  # Font size for "Category B"
        venn.get_label_by_id('C').set_fontsize(12)  # Font size for "Category C"

        # Adjust the font size of subset labels (the numbers inside the circles)
        for text in venn.subset_labels:
            if text:  # Check if the label exists (some subsets might be None)
                text.set_fontsize(users_settings['font_size'])

        # Display the plot in Streamlit
        st.pyplot(plt.gcf())
        plt.close()

    elif n_categories == 1:
        st.error(f'Venn diagrams only work with 2-3 categories. The metadata {selected_metadata} has one category.')
    else:
        st.error(
            f'Venn diagrams only work with 2-3 categories. The metadata {selected_metadata} has {n_categories} categories.')






import pandas as pd
from pandas import DataFrame
import numpy as np
from pathlib import Path

# read_table_xlsx = '/Users/tillmacher/Documents/GitHub/TaxonTableTools/_tutorial_files/tutorial_read_table_TTT.xlsx'
# taxonomy_table_xlsx = '/Users/tillmacher/Documents/GitHub/TaxonTableTools/_tutorial_files/tutorial_taxonomy_table.xlsx'
# TaXon_table_name = 'this_is_a_test'
# taxonomy_table_format = 'JAMP hit'
# path_to_outdirs = '/Users/tillmacher/Desktop/ttt_projects/Projects/Dessau_Vertebrates'

def taxon_table_converter_ttt(read_table_xlsx, taxonomy_table_xlsx, TaXon_table_name, taxonomy_table_format, path_to_outdirs):

    ## format working sheet name
    if taxonomy_table_format == 'Taxonomy table (APSCALE)':
        taxonomy_table_format = 'Taxonomy table'

    ## collect both input files
    taxonomy_table_xlsx =  Path(taxonomy_table_xlsx)
    read_table_xlsx = Path(read_table_xlsx)

    #ä create filename and path for output file
    Output_name = TaXon_table_name + ".xlsx"
    Output_file = Path(path_to_outdirs) / "TaXon_tables" / Output_name

    ## store the file name for later use
    file_name = taxonomy_table_xlsx.name

    ## create dataframes for both files
    taxonomy_df = pd.read_excel(taxonomy_table_xlsx, taxonomy_table_format, header=0)
    if taxonomy_table_format == "BOLDigger hit":
        taxonomy_df = taxonomy_df.drop(columns=['Flags'])
    read_table_df = pd.read_excel(read_table_xlsx, header=0)

    ## create a new dataframe
    TaXon_table_df = taxonomy_df

    # check if all OTU are correctly sorted and present in both files
    if taxonomy_df["ID"].to_list() == read_table_df["ID"].to_list():

        ## append the sequences to the TaXon stable
        TaXon_table_df["seq"] = read_table_df["Sequences"]

        ## remove the sequence column from the read table
        read_table_df.drop('Sequences', axis='columns', inplace=True)

        ## remove the ID column from the read table
        read_table_df.drop('ID', axis='columns', inplace=True)

        ## add samples to the dataframe
        TaXon_table_df = pd.concat([TaXon_table_df, read_table_df], axis=1)

        ## check if species are present as "Genus" + "Epithet"
        new_species_column = []
        for OTU in TaXon_table_df[["Genus", "Species"]].fillna("nan").values.tolist():
            if (OTU != ["nan", "nan"] and OTU[1] != 'nan'):
                if OTU[0] not in OTU[1]:
                    new_species_column.append(OTU[0] + " " + OTU[1])
                else:
                    new_species_column.append(OTU[1])
            else:
                new_species_column.append("")

        ## add new species column to the dataframe
        TaXon_table_df["Species"] = new_species_column

        ## save the newly created Taxon table in TaXon format as excel file
        TaXon_table_df.to_excel(Output_file, sheet_name='TaXon table', index=False)

        return ('Finished: Please select your new TaXon table above.')

    else:
        return 'Error: Please check your input files!'

def taxon_table_converter_qiime2(read_table_tsv, taxonomy_table_xlsx, TaXon_table_name, taxonomy_table_format, path_to_outdirs):

    taxonomy_table_xlsx =  Path(taxonomy_table_xlsx)
    read_table_tsv = Path(read_table_tsv)

    # create filename and path for output file
    Output_name = TaXon_table_name + ".xlsx"
    Output_file = Path(path_to_outdirs) / "TaXon_tables" / Output_name

    # store the file name for later use
    file_name = taxonomy_table_xlsx.name

    # create datafrmes for both files
    taxonomy_df = pd.read_excel(taxonomy_table_xlsx, taxonomy_table_format, header=0)
    if taxonomy_table_format == "BOLDigger hit":
        taxonomy_df = taxonomy_df.drop(columns=['Flags'])

    read_table_df = pd.read_csv(Path(read_table_tsv), sep="\t")

    # drop the first row
    read_table_df = read_table_df.iloc[1:]
    read_table_df = read_table_df.reset_index(drop=True)

    ## create a new dataframe
    TaXon_table_df = taxonomy_df

    # check if all OTU are correctly sorted and present in both files
    if taxonomy_df["ID"].to_list() == read_table_df["id"].to_list():

        ## append the sequences to the TaXon stable
        TaXon_table_df["seq"] = read_table_df["Sequence"].values.tolist()

        ## remove the sequence column from the read table
        read_table_df.drop('Sequence', axis='columns', inplace=True)

        ## remove the ID column from the read table
        read_table_df.drop('id', axis='columns', inplace=True)

        ## add samples to the dataframe
        TaXon_table_df = pd.concat([TaXon_table_df, read_table_df], axis=1)

        ## check if species are present as "Genus" + "Epithet"
        new_species_column = []
        for OTU in TaXon_table_df[["Genus", "Species"]].fillna("nan").values.tolist():
            if (OTU != ["nan", "nan"] and OTU[1] != 'nan'):
                if OTU[0] not in OTU[1]:
                    new_species_column.append(OTU[0] + " " + OTU[1])
                else:
                    new_species_column.append(OTU[1])
            else:
                new_species_column.append("")

        ## add new species column to the dataframe
        TaXon_table_df["Species"] = new_species_column

        ## save the newly created Taxon table in TaXon format as excel file
        TaXon_table_df.to_excel(Output_file, sheet_name='TaXon table', index=False)

        return ('Finished: Please select your new TaXon table above.')

    else:
        return 'Error: Please check your input files!'



import pandas as pd
import math
import numpy as np
from taxontabletools2.utilities import collect_traits
from taxontabletools2.utilities import strip_traits
from taxontabletools2.utilities import collect_metadata
from taxontabletools2.utilities import load_df
from taxontabletools2.utilities import collect_replicates
from taxontabletools2.utilities import export_taxon_table
from taxontabletools2.utilities import update_taxon_table
import requests, datetime, json, time
from stqdm import stqdm

def gbif_accession(taxon_table_xlsx, taxon_table_df, samples, metadata_df, traits_df, higherTaxonKey):
    ## create copies of the dataframes
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    ## store results here
    species_dict = {'': ['', '', '', '', '']}
    OTU_species = []
    OTU_keys = []
    OTU_genusKeys = []
    OTU_synonyms = []
    OTU_iucnRedListCategory = []
    OTU_habitat = []

    for OTU in stqdm(taxon_table_df.values.tolist()):
        species = OTU[6]

        ## request GBIF
        if species != '':
            if species not in species_dict.keys():
                query = species.replace(' ', '%20')
                # Initialize a counter for the number of attempts
                attempts = 0
                # Initialize a flag for whether the request was successful
                success = False

                while not success and attempts < 20:
                    try:
                        # Make a GET request to the API
                        response = requests.get(f'https://api.gbif.org/v1/species/match?name={query}')

                        # If the request was successful, parse the response text as JSON
                        if response.status_code == 200:
                            data = json.loads(response.text)
                            success = True
                        else:
                            attempts += 1
                    except requests.exceptions.RequestException as e:
                        # If there was a network problem (e.g. DNS resolution, refused connection, etc), increment the counter and try again
                        attempts += 1

                # If the request was not successful after 20 attempts, print a warning
                if not success:
                    species_dict[species] = [''] * 6
                else:
                    # add data to dict
                    species_dict[species] = [data.get('species', ''), data.get('speciesKey', ''),
                                             data.get('genusKey', ''), data.get('synonym', '')]

                    # Make a GET request to the API to get IUCN Red List Category
                    response = requests.get(
                        f"https://api.gbif.org/v1/species/{data.get('speciesKey', '')}/iucnRedListCategory")
                    if response.status_code == 200:
                        iucn_data = json.loads(response.text)
                        species_dict[species].append(iucn_data.get('code', ''))
                    else:
                        species_dict[species].append('')

                    # Make a GET request to the API to get habitat
                    response = requests.get(
                        f"https://api.gbif.org/v1/species/{data.get('speciesKey', '')}/speciesProfiles")
                    if response.status_code == 200:
                        habitat_data = json.loads(response.text)
                        habitat = ', '.join([j for j in sorted(set([i.get('habitat', '').lower() for i in habitat_data['results']])) if j != ''])
                        species_dict[species].append(habitat)
                    else:
                        species_dict[species].append('')

                time.sleep(0.5)

            ## append results
            OTU_species.append(species_dict[species][0])
            OTU_keys.append(str(species_dict[species][1]))
            OTU_genusKeys.append(str(species_dict[species][2]))
            OTU_synonyms.append(str(species_dict[species][3]))
            OTU_iucnRedListCategory.append(species_dict[species][4])
            OTU_habitat.append(species_dict[species][5])
        else:
            ## append results
            OTU_species.append('')
            OTU_keys.append('')
            OTU_genusKeys.append('')
            OTU_synonyms.append('')
            OTU_iucnRedListCategory.append('')
            OTU_habitat.append('')

    ## append to traits df
    traits_df['speciesKey'] = OTU_keys
    traits_df['genusKey'] = OTU_genusKeys
    traits_df['GBIF species'] = OTU_species
    traits_df['Synonym'] = OTU_synonyms
    traits_df['iucnRedListCategory'] = OTU_iucnRedListCategory
    traits_df['habitat'] = OTU_habitat

    ## export table
    suffix = 'test'
    update_taxon_table(taxon_table_xlsx, taxon_table_df, traits_df, metadata_df, suffix)


taxon_table_xlsx = '/Users/tillmacher/Desktop/TTT_projects/Projects/AA_test/TaXon_tables/Elbe_tele02_taxon_table_merged_NCsub_filtered_norm_fish.xlsx'
taxon_table_df = pd.read_excel(taxon_table_xlsx).fillna('')
metadata_df = collect_metadata(taxon_table_df)
traits_df = collect_traits(taxon_table_df)
taxon_table_df = strip_traits(taxon_table_df)
samples = taxon_table_df.columns.tolist()[10:]
higherTaxonKey= 1



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="taxontabletools2", # Replace with your own username
    version="2.0.0",
    author="Till-Hendrik Macher",
    author_email="macher@uni-trier.de",
    description="taxontabletools - A comprehensive, platform-independent graphical user interface software to explore and visualise DNA metabarcoding data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/taxontabletools2",
    packages=setuptools.find_packages(),
    license = 'MIT',
    install_requires = [
            'matplotlib == 3.6.3',
            'matplotlib_venn == 0.11.7',
            'numpy == 2.1.3',
            'pandas == 2.2.3',
            'plotly == 5.9.0',
            'psutil == 5.9.4',
            'pymannkendall == 1.4.3',
            'Requests == 2.32.3',
            'scipy == 1.14.1',
            'seqconverter == 3.1.0',
            'statsmodels == 0.13.5',
            'stqdm == 0.0.5',
            'streamlit == 1.25.0',
            'update_checker == 0.18.0'
    ],
    include_package_data = True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
    entry_points = {
        "console_scripts" : [
            "taxontabletools2 = taxontabletools2.__main__:main",
        ]
    },
)



import streamlit as st
import pandas as pd
import plotly.graph_objects as go
from pathlib import Path
import statistics

# TaXon_table_df = pd.read_excel('/Users/tillmacher/Desktop/TTT_projects/Projects/AA_test/TaXon_tables/tutorial_taxon_table.xlsx').fillna('nan')
# settings = {'path_to_projects': Path('/Users/tillmacher/Desktop/ttt_projects/Projects'), 'template': 'plotly', 'color1': 'blue', 'color2': 'black', 'colorsequence': 'Plotly', 'colorscale': 'jet', 'font_size': 16, 'clustering_unit': 'OTUs'}


@st.cache_data()
def basic_stats_reads(settings, TaXon_table_df):

    """ Calculate the number of reads per sample """
    samples = TaXon_table_df.columns.tolist()[10:]
    y_values = {i:sum([j for j in TaXon_table_df[i].values.tolist() if j != 0]) for i in samples}
    y_values = dict(sorted(y_values.items(), reverse=True, key=lambda item: item[1]))
    x_values = list(y_values.keys())

    max_reads, min_reads, avg_reads, stdev = max(y_values.values()), min(y_values.values()), round(statistics.mean(y_values.values()), 2), round(statistics.stdev(y_values.values()), 2)

    fig = go.Figure()
    fig = fig.add_trace(go.Bar(x=x_values,y=list(y_values.values())))
    fig.update_yaxes(title = 'Reads', title_font=dict(size=settings['font_size']), tickfont=dict(size=settings['font_size']))
    fig.update_xaxes(title_font=dict(size=settings['font_size']), tickfont=dict(size=settings['font_size']))
    fig.update_layout(template=settings['template'], font_size=settings['font_size'])

    return fig, min_reads, max_reads, avg_reads, stdev

@st.cache_data()
def basic_stats_OTUs(settings, TaXon_table_df):

    """ Calculate the number of OTUs per sample """

    samples = TaXon_table_df.columns.tolist()[10:]
    y_values = {i:len([j for j in TaXon_table_df[i].values.tolist() if j != 0]) for i in samples}
    y_values = dict(sorted(y_values.items(), reverse=True, key=lambda item: item[1]))
    x_values = list(y_values.keys())

    max_OTUs, min_OTUs, avg_OTUs = max(y_values.values()), min(y_values.values()), round(statistics.mean(y_values.values()), 2)

    fig = go.Figure()
    fig = fig.add_trace(go.Bar(x=x_values,y=list(y_values.values())))
    fig.update_layout(template=settings['template'], font_size=settings['font_size'])
    fig.update_yaxes(title = settings['clustering_unit'], title_font=dict(size=settings['font_size']), tickfont=dict(size=settings['font_size']))
    fig.update_xaxes(title_font=dict(size=settings['font_size']), tickfont=dict(size=settings['font_size']))

    return fig, min_OTUs, max_OTUs, avg_OTUs

@st.cache_data()
def taxonomic_richness(settings, TaXon_table_df):

    """ Calculate the taxonomic richness """

    taxonomic_levels = ["Phylum", "Class", "Order", "Family", "Genus", "Species"]

    statistics_list, statistics_set, statistics_dict, highest_level_dict = [], [], {}, {}

    for taxon_to_evaluate in taxonomic_levels:
        taxa_list = [x for x in TaXon_table_df[taxon_to_evaluate].values.tolist() if str(x) != 'nan']
        statistics = taxon_to_evaluate, len(taxa_list)
        statistics_set.append(len(set(taxa_list)))
        statistics_list.append(list(statistics))
        statistics_dict[taxon_to_evaluate] = len(taxa_list)

    highest_level_dict["Phylum"] = statistics_dict["Phylum"] - statistics_dict["Class"]
    highest_level_dict["Class"] = statistics_dict["Class"] - statistics_dict["Order"]
    highest_level_dict["Order"] = statistics_dict["Order"] - statistics_dict["Family"]
    highest_level_dict["Family"] = statistics_dict["Family"] - statistics_dict["Genus"]
    highest_level_dict["Genus"] = statistics_dict["Genus"] - statistics_dict["Species"]
    highest_level_dict["Species"] = statistics_dict["Species"]

    taxon_levels = list(highest_level_dict.keys())
    number_of_taxa_per_level = statistics_set

    fig = go.Figure(data=[go.Bar(x=taxon_levels, y=number_of_taxa_per_level, name="Taxon", textposition="outside", cliponaxis=False, text=number_of_taxa_per_level)])
    fig.update_layout(template=settings['template'], font_size=settings['font_size'])
    fig.update_yaxes(title = '# taxa', title_font=dict(size=settings['font_size']), tickfont=dict(size=settings['font_size']))
    fig.update_xaxes(title_font=dict(size=settings['font_size']), tickfont=dict(size=settings['font_size']))

    return fig

@st.cache_data()
def taxonomic_resolution(settings, TaXon_table_df):
    """Calculate the taxonomic richness."""

    # All species
    all_species = TaXon_table_df['Species'].values.tolist()
    unique_species = sorted(set(all_species))
    species_count = {i:all_species.count(i) for i in unique_species if i != ''}
    species_count = dict(sorted(species_count.items(), key=lambda item: item[1], reverse=True))
    x_values = list(species_count.keys())[:15]
    display_x_values = []
    for value in x_values:
        try:
            display_x_values.append(f"<i>{value.split(' ')[0][0]}. {value.split(' ')[1]}<i>")
        except:
            display_x_values.append(value)

    y_values = list(species_count.values())[:15]

    # Title setup (not used in current code but retained for completeness)
    title = f"# {settings['clustering_unit']}"

    # Create bar chart
    fig = go.Figure(data=go.Bar(x=display_x_values, y=y_values))

    fig.update_layout(
        template=settings['template'],
        font_size=settings['font_size'],
        yaxis_title=settings['clustering_unit'],
    )

    fig.update_yaxes(title_font=dict(size=settings['font_size']), tickfont=dict(size=settings['font_size']))
    fig.update_xaxes(title_font=dict(size=settings['font_size']), tickfont=dict(size=settings['font_size']))

    return fig

#



import subprocess
import sys
from pathlib import Path
import pandas as pd
import tkinter as tk
from tkinter import filedialog, messagebox
from update_checker import update_check
import importlib.metadata
import threading
import time

## Check for updates
def check_package_update(package_name):
    try:
        # Get the currently installed version
        installed_version = importlib.metadata.version(package_name)
    except importlib.metadata.PackageNotFoundError:
        return f"{package_name} is not installed."

    # Check for updates
    res = update_check(package_name, installed_version)
    if res:
        return f"New version of {package_name} is available: {res['latest_version']} (Installed: {installed_version})"
    else:
        return f"{package_name} is up to date (Version: {installed_version})."

# Function to select a folder
def select_folder():
    global path_to_outdirs
    new_folder = filedialog.askdirectory(title="Select the projects folder")
    if new_folder:
        path_to_outdirs = new_folder
        folder_label.config(text=f"Selected Folder: {path_to_outdirs}")
    else:
        folder_label.config(text=f"Using previous folder: {path_to_outdirs}")

# Function to run the Streamlit app
def run_streamlit_app():
    try:
        print('Press CTRL + C to close APSCALE!')
        subprocess.call(['streamlit', 'run', './taxontabletools_2.0.py', '--theme.base', 'light' ])
    except KeyboardInterrupt:
        sys.exit()

# Function to close the tkinter app and continue with the script
def start_app():
    if not path_to_outdirs:
        messagebox.showwarning("Warning", "Please select a projects folder before continuing.")
        return

    # Save the selected folder to user_preferences.xlsx
    save_folder_to_preferences()

    # Hide window
    root.withdraw()

    # Run the Streamlit app in a new thread to avoid freezing the GUI
    threading.Thread(target=run_streamlit_app).start()

# Function to save the selected folder to the preferences file
def save_folder_to_preferences():
    user_preferences_df.loc[user_preferences_df['Variable'] == 'path_to_outdirs', 'Value'] = path_to_outdirs
    user_preferences_df.to_excel(user_preferences_xlsx, index=False)

# Tkinter window setup
root = tk.Tk()
root.title("TaxonTableTools Start Window")
root.geometry("400x200")

# Check for APSCALE package update
update_info = check_package_update('taxontabletools')

# Update label
update_label = tk.Label(root, text=update_info, wraplength=300, justify="left")
update_label.pack(pady=10)

# Paths and user preferences setup
path_to_ttt = Path(__file__).resolve().parent
user_preferences_xlsx = path_to_ttt.joinpath('user_preferences.xlsx')

# Read user preferences from the file
user_preferences_df = pd.read_excel(user_preferences_xlsx).fillna('')
path_to_outdirs = user_preferences_df.loc[user_preferences_df['Variable'] == 'path_to_outdirs', 'Value'].values[0]

# Display the current folder
folder_label = tk.Label(root, text=f"Current folder: {path_to_outdirs if path_to_outdirs else 'No folder selected'}")
folder_label.pack(pady=10)

# Button to select a new folder
select_button = tk.Button(root, text="Select Projects Folder", command=select_folder)
select_button.pack(pady=10)

# Start button to continue with the script
start_button = tk.Button(root, text="Start", command=start_app)
start_button.pack(pady=10)

# Run tkinter main loop
root.mainloop()


import pandas as pd
import numpy as np
from pathlib import Path
from taxontabletools2.utilities import filter_taxontable
import plotly.graph_objects as go
import plotly.express as px
from scipy.spatial.distance import pdist, squareform

def alpha_boxplot(path_to_outdirs, taxon_table_xlsx, taxon_table_df, samples, metadata_df, selected_metadata,
                   traits_df, selected_traits, user_settings, tool_settings):

    # Make copies of input dataframes to prevent altering the originals
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    # Extract relevant settings from the tool settings
    selected_metadata = tool_settings['selected_metadata']
    taxonomic_level = tool_settings['taxonomic_level']

    # Get unique taxa at the specified taxonomic level and exclude empty values
    all_taxa = taxon_table_df[taxonomic_level].dropna().unique()

    # Initialize result dictionary
    res = {}
    if selected_metadata == 'All samples':
        for sample in samples:
            res[sample] = len([i for i in taxon_table_df[[taxonomic_level, sample]].groupby(taxonomic_level).sum().loc[lambda x: x[sample] > 0].index.tolist() if i != ''])

    else:
        # Case where specific categories from metadata are selected
        categories = [i for i in metadata_df[selected_metadata].dropna().unique() if i != '']

        # Filter metadata to keep relevant columns only
        metadata_df = metadata_df[['Sample', selected_metadata]]

        for category in categories:
            # Get the samples belonging to the current category
            category_samples = metadata_df.loc[metadata_df[selected_metadata] == category, 'Sample'].tolist()
            category_list = []
            for sample in category_samples:
                category_list.append(len([i for i in taxon_table_df[[taxonomic_level, sample]].groupby(taxonomic_level).sum().loc[lambda x: x[sample] > 0].index.tolist() if i != '']))
            res[category] = category_list

    # Use color map based on user settings
    colorscale_name = user_settings.get('colorsequence', 'Plotly')
    colors = getattr(px.colors.qualitative, colorscale_name, px.colors.qualitative.Plotly)*100
    color=0

    # Initialize and populate the figure
    fig = go.Figure()
    for key, values in res.items():
        if str(type(values)) != "<class 'list'>":
            values = [values]
        fig.add_trace(go.Box(y=values, name=key, marker_color=colors[color]))
        color+=1

    # Update layout
    fig.update_yaxes(rangemode='tozero')
    fig.update_layout(barmode='stack',
                      title=f'Alpha diversity: {taxonomic_level}, {selected_metadata}',
                      template=user_settings['template'],
                      font_size=user_settings['font_size'],
                      width=user_settings['plot_width'],
                      height=user_settings['plot_height'],
                      showlegend=user_settings['show_legend'],
                      yaxis_title=f'Number of Taxa ({taxonomic_level})',
                      yaxis_title_font=dict(size=user_settings['font_size']),
                      yaxis_tickfont=dict(size=user_settings['font_size']),
                      xaxis_title_font=dict(size=user_settings['font_size']),
                      xaxis_tickfont=dict(size=user_settings['font_size']))

    return fig


def distance_matrix(path_to_outdirs, taxon_table_xlsx, taxon_table_df, samples, metadata_df, selected_metadata,
                    traits_df, selected_traits, user_settings, tool_settings):
    # Make copies of input dataframes to prevent altering the originals
    taxon_table_df = taxon_table_df.copy()
    metadata_df = metadata_df.copy()
    traits_df = traits_df.copy()

    # Extract relevant settings from the tool settings
    selected_metadata = tool_settings['selected_metadata']
    taxonomic_level = tool_settings['taxonomic_level']
    metric = tool_settings['metric']

    # Get unique taxa at the specified taxonomic level and exclude empty values
    all_taxa = [i for i in taxon_table_df[taxonomic_level].dropna().unique() if i != '']

    # Initialize result dictionary
    res = {}

    if selected_metadata == 'All samples':
        # Case where all samples are used
        for sample in samples:
            # Sum the reads for each taxon in the current sample and filter for non-zero values
            sample_taxa = taxon_table_df[[taxonomic_level, sample]].groupby(taxonomic_level).sum()
            sample_taxa = sample_taxa.loc[sample_taxa[sample] > 0, sample]

            # Store binary presence/absence data if metric is 'Jaccard', otherwise store actual values
            if metric == 'Jaccard':
                res[sample] = [1 if taxon in sample_taxa.index else 0 for taxon in all_taxa]
            else:
                res[sample] = [sample_taxa[taxon] if taxon in sample_taxa.index else 0 for taxon in all_taxa]

        # Convert result dict into DataFrame (rows: taxa, columns: samples)
        presence_absence_df = pd.DataFrame(res, index=all_taxa).transpose()

        # Calculate distance between samples
        if metric == 'Jaccard':
            distances = pdist(presence_absence_df, metric=metric)
        else:
            distances = pdist(presence_absence_df, metric=metric)

        # Convert to square form distance matrix
        distance_matrix = squareform(distances)

        # Convert to a DataFrame for easier manipulation
        distance_matrix_df = pd.DataFrame(distance_matrix, index=samples, columns=samples)

    else:
        # Case where specific metadata categories are selected
        categories = [i for i in metadata_df[selected_metadata].dropna().unique() if i != '']

        # Filter metadata to keep relevant columns
        metadata_df = metadata_df[['Sample', selected_metadata]]

        for category in categories:
            # Get the samples belonging to the current category
            category_samples = metadata_df.loc[metadata_df[selected_metadata] == category, 'Sample'].tolist()

            # Sum reads across all samples in category for each taxon
            category_taxa = taxon_table_df[[taxonomic_level] + category_samples].groupby(taxonomic_level).sum()
            category_taxa = category_taxa.loc[category_taxa[category_samples].sum(axis=1) > 0].sum(axis=1)

            # Store binary presence/absence data if metric is 'Jaccard', otherwise store actual values
            if metric == 'Jaccard':
                res[category] = [1 if taxon in category_taxa.index else 0 for taxon in all_taxa]
            else:
                res[category] = [category_taxa[taxon] if taxon in category_taxa.index else 0 for taxon in all_taxa]

        # Convert result dict into DataFrame (rows: taxa, columns: categories)
        presence_absence_df = pd.DataFrame(res, index=all_taxa).transpose()

        # Calculate distance between categories
        if metric == 'Jaccard':
            distances = pdist(presence_absence_df, metric=metric)
        else:
            distances = pdist(presence_absence_df, metric=metric)

        # Convert to square form distance matrix
        distance_matrix = squareform(distances)

        # Convert to a DataFrame for easier manipulation
        distance_matrix_df = pd.DataFrame(distance_matrix, index=categories, columns=categories)


    # Use color map based on user settings
    colorscale_name = user_settings.get('colorscale', 'Viridis')

    # Initialize heatmap figure
    fig = go.Figure(data=go.Heatmap(
        z=distance_matrix_df.values.tolist(),
        x=distance_matrix_df.columns.tolist(),
        y=distance_matrix_df.index.tolist(),
        zmin=0,
        zmax=1,
        hoverongaps=False,
        colorscale=colorscale_name))

    # Update layout
    fig.update_xaxes(dtick='linear')
    fig.update_yaxes(dtick='linear')
    fig.update_layout(title=f'Beta diversity: {taxonomic_level}, {metric}',
                      template=user_settings['template'],
                      font_size=user_settings['font_size'],
                      width=user_settings['plot_width'],
                      height=user_settings['plot_height'],
                      showlegend=user_settings['show_legend'],
                      yaxis_title_font=dict(size=user_settings['font_size']),
                      yaxis_tickfont=dict(size=user_settings['font_size']),
                      xaxis_title_font=dict(size=user_settings['font_size']),
                      xaxis_tickfont=dict(size=user_settings['font_size']))

    return fig


